import tkinter as tk
from viewable import Viewable
from exn.view.top import Top
from exn.view.board import Board
from exn.view.footer import Footer
from exonote import split_target
from viewstack import ViewStack


class Front(Viewable):
    def __init__(self, context, target):
        super().__init__()
        self._context = context
        self._target = target
        self._manager = context.manager
        self._dao = context.manager.dao
        self._app = context.manager.app
        self._top = None
        self._board = None
        self._footer = None
        self._viewstack = None
        self._board_container = None
        self._new = True

    @property
    def manager(self):
        return self._manager

    @property
    def filename(self):
        return self._filename

    @property
    def top(self):
        return self._top

    @property
    def board(self):
        return self._board

    @property
    def footer(self):
        return self._footer

    def open(self, target):
        filename, sid = split_target(target)
        if filename in self._dao.get_blocklist():
            return False
        self._filename = filename
        self._board = self._viewstack.views.get(filename)
        if self._board:
            self._viewstack.lift(filename)
        else:
            old_selection = self._viewstack.selection
            self._board = Board(self, filename)
            self._viewstack.add(filename, self._board)
            if self._board.failed:
                self._viewstack.destroy(filename)
                if old_selection:
                    filename = old_selection.name
                    self._board = self._viewstack.lift(filename)
                    self._filename = filename
                return False
        if sid:
            self._board.viewer.goto(sid)
        self._dao.update_history(self._filename)
        self._update_app_title(filename)
        self._footer.update_page_status(filename)
        self._footer.finder.disable()
        return True

    def refresh(self):
        if not self._filename:
            return
        filename = self._filename
        if self._filename not in self._viewstack.views:
            return
        self._viewstack.destroy(filename)
        self.open(filename)

    def _create_body(self, parent):
        return tk.Frame(parent, name="front")

    def _build(self):
        #self.body.bind("<Visibility>", self._handle_visibility_event)
        # top
        self._top = Top(self)
        self._top.build_pack(self.body, fill=tk.X)
        # board container
        self._board_container = tk.Frame(self.body, name="board_stack")
        self._board_container.pack(fill=tk.BOTH, expand=True)
        self._viewstack = ViewStack(self._board_container)
        #self._board = Board(self)
        #self._board.build_pack(self.body, fill=tk.BOTH, expand=True)
        # footer
        on_leave_focus = lambda: self._board.editor.focus_set()
        self._footer = Footer(self, on_leave_focus=on_leave_focus)
        self._footer.build_pack(self.body, fill=tk.X)
        self.open(self._target)

    def _on_map(self):
        pass

    def _on_remap(self):
        #print(event.state)
        #if event.state != "VisibilityUnobscured":
        #    return
        if not self._board:
            return
        self._board.viewer.editor.focus_set()

    def _on_destroy(self):
        if not self._board:
            return
        images = self._board.viewer.data.get("images")
        if not images:
            return
        images[:] = []

    def _update_app_title(self, filename):
        for item in self._manager.index:
            if item.filename == filename:
                cache = item.title if item.title else "- No Title -"
                #title = "{}   |   {}   |   Exonote Reader".format(cache, filename)
                title = "{}   -   Exonote Reader".format(cache)
                self._app.root.title(title)



import os
import os.path
import sys
from exn import constant, utils
from exonote import Style, split_target, get_callable
from gaspium import App
from exn.view import Front
from exn.dao import Dao
from exn.theme import Theme, Constants as theme_const, get_standard_font_family


class Manager:
    def __init__(self, dossier, restriction):
        self._dossier = dossier
        self._restriction = restriction
        self._app = None
        self._dao = Dao(dossier)
        self._app_theme = None
        self._embedded_gui_theme = None
        self._viewer_style = None
        self._index = None
        self._setup()

    @property
    def app(self):
        return self._app

    @property
    def dossier(self):
        return self._dossier

    @property
    def restriction(self):
        return self._restriction

    @property
    def index(self):
        return self._index

    @property
    def dao(self):
        return self._dao

    @property
    def app_theme(self):
        return self._app_theme

    @property
    def viewer_style(self):
        return self._viewer_style

    @property
    def embedded_gui_theme(self):
        return self._embedded_gui_theme

    def start(self, target=None):
        if not target:
            target = self._get_default_note()
            if not target:
                return
        filename, sid = split_target(target)
        if filename in self._dao.get_blocklist():
            msg = "'{}' appears in the blocklist !"
            print(msg.format(filename))
            return
        #self._dao.update_history(filename)
        self._app = App(name="exn", title="Exonote Reader",
                        manager=self, geometry="700x600",
                        remember_geometry_change=True,
                        resizable=(True, True),
                        caching=True, show_page_title=False,
                        navbar=None, page_from_cli=False)
        self._app.root.minsize(*constant.WINDOW_MINSIZE)
        self._app_theme(self._app.root)
        self._update_theme_const()
        kwargs = {"target": target}
        self._app.attach(Front, pid="front", kwargs=kwargs)
        self._app.open("front")
        self._app.start()

    def _setup(self):
        # update sys.path
        sys.path.insert(1, self._dossier)
        # default themes
        self._app_theme = Theme
        #self._embedded_gui_theme = Theme2
        # default viewer style
        self._viewer_style = Style()
        # load user defined style
        data = self._load_user_defined_style()
        # update app and embedded gui themes
        self._update_themes(data)
        # update viewer style
        self._update_viewer_style(data)
        # get index data
        self._load_index()

    def _load_user_defined_style(self):
        # read style file
        style_file = os.path.join(self._dossier, ".exn", "style")
        return utils.read_style_file(style_file)

    def _update_themes(self, data):
        if not data:
            return
        # get app theme
        app_theme = data.get("app_theme")
        if app_theme:
            cache = get_callable(app_theme)
            if cache:
                self._app_theme = cache[1]
        # get app theme
        embedded_gui_theme = data.get("embedded_gui_theme")
        if embedded_gui_theme:
            # get app theme
            cache = get_callable(embedded_gui_theme)
            if cache:
                self._embedded_gui_theme = cache[1]

    def _update_viewer_style(self, data):
        if not data:
            return
        valid_options = {"horizontal_margin", "vertical_margin",
                         "foreground_color", "background_color",
                         "selection_background_color", "highlight_color",
                         "active_highlight_color", "spacing", "font_family",
                         "font_size", "codeblock_foreground_color",
                         "heading_foreground_color", "link_foreground_color",
                         "executable_link_foreground_color", "code_background_color",
                         "monospace_font_family", "warning_color", "notice_color"}
        integers_options = {"horizontal_margin", "vertical_margin",
                            "spacing", "font_size"}
        for option in valid_options:
            value = data.get(option, "")
            if not value:
                continue
            if option in integers_options:
                try:
                    value = int(value)
                except ValueError as e:
                    continue
            setattr(self._viewer_style, option, value)

    def _get_default_note(self):
        history = self._dao.get_history()
        if not history:
            index_path = os.path.join(self._dossier, "index")
            index = utils.IndexParser.parse(index_path)
            if not index:
                return
            return index[0].filename
        return history[0]

    def _load_index(self):
        path = os.path.join(self._dossier, "index")
        self._index = utils.IndexParser.parse(path)
        self._index = self._index if self._index else list()

    def _update_theme_const(self):
        theme_const.FONT_FAMILY = get_standard_font_family(name="TkTextFont")
        theme_const.FONT_FAMILY_MONO = get_standard_font_family(name="TkFixedFont")



import tkinter as tk
from viewable import Viewable
from exonote import Viewer
from exonote.editor import Editor
from exn.view.switcher import Switcher


class Board(Viewable):
    def __init__(self, front, filename):
        super().__init__()
        self._front = front
        self._filename = filename
        self._editor = None
        self._viewer = None
        self._toc_frame = None
        self._toc = None
        self._toc_is_visible = False
        self._failed = False

    @property
    def front(self):
        return self._front

    @property
    def editor(self):
        return self._editor

    @property
    def viewer(self):
        return self._viewer

    @property
    def failed(self):
        return self._failed

    def _create_body(self, parent):
        return tk.Frame(parent)

    def _build(self):
        self._toc_frame = tk.Frame(self.body)
        self._toc_frame.pack(side=tk.LEFT, fill=tk.Y)
        self._editor = Editor(self.body, width=0, height=0,
                              takefocus=False)
        self._editor.vbar.config(takefocus=False)
        self._editor.pack(side=tk.RIGHT, fill=tk.BOTH, expand=True)
        #style = self._reader.style
        #inserter = self._reader.inserter
        restriction = self._front.manager.restriction
        self._viewer = Viewer(self._editor, restriction=restriction,
                              style=self._front.manager.viewer_style,
                              on_press_left=self._open_prev_page,
                              on_press_right=self._open_next_page,
                              update_sys_path=False)
        if not self._viewer.open(self._filename):
            self._failed = True
            return
        on_open = lambda viewer, filename: self._open_page(filename)
        self._viewer.on_open = on_open

    def _on_map(self):
        self._bind_events_handlers()
        self._editor.focus_set()

    def _on_remap(self):
        self._editor.focus_set()

    def _bind_events_handlers(self):
        top = self._front.top
        # refresh
        self._editor.bind("<F5>", lambda e: self._front.refresh())
        # home
        self._editor.bind("<h>", lambda e: top.click_home())
        self._editor.bind("<H>", lambda e: top.click_home())
        # toc
        self._editor.bind("<t>", lambda e: top.click_toc())
        self._editor.bind("<T>", lambda e: top.click_toc())
        # mark
        self._editor.bind("<i>", lambda e: top.click_info())
        self._editor.bind("<I>", lambda e: top.click_info())
        # goto
        self._editor.bind("<g>", lambda e: top.click_goto())
        self._editor.bind("<G>", lambda e: top.click_goto())
        # past
        self._editor.bind("<p>", lambda e: top.click_past())
        self._editor.bind("<P>", lambda e: top.click_past())
        # search
        self._editor.bind("<s>", lambda e: top.click_search())
        self._editor.bind("<S>", lambda e: top.click_search())
        # about
        self._editor.bind("<a>", lambda e: top.click_about())
        self._editor.bind("<A>", lambda e: top.click_about())
        # find
        self._editor.bind("<f>", lambda e: self._on_press_key_f())
        self._editor.bind("<F>", lambda e: self._on_press_key_f())
        # switcher controls
        self._editor.bind("<Control-KeyPress-Tab>", self._open_switcher)

    def _open_switcher(self, e):
        switcher = Switcher(self._front)
        switcher.build(self.body)
        return "break"

    def _on_press_key_f(self):
        self._front.footer.enable_finder()

    def _open_page(self, filename):
        self._front.open(filename)

    def _open_prev_page(self):
        footer = self._front.footer
        if not footer:
            return
        footer.open_prev_page()

    def _open_next_page(self):
        footer = self._front.footer
        if not footer:
            return
        footer.open_next_page()



import os
import os.path
import importlib
import jesth
from operator import itemgetter
from exonote import IndexParser, dto


# no usage (for the moment)
def load_extensions(context):
    # context = dto.ExtensionContext(dossier_path, app_instance, manager)
    dossier = os.getcwd()
    extensions_filename = os.path.join(dossier, ".exn", "extensions")
    if not os.path.isfile(extensions_filename):
        return None
    doc = jesth.read(extensions_filename)
    section = doc.get("")
    body = section.body if section else list()
    if not body:
        return None
    for line in body:
        if line.startswith("#"):
            continue
        if not line:
            continue
        cache = line.split(":")
        if len(cache) > 2:
            continue
        if len(cache) == 1:
            module_name, callable_name = cache[0], "main"
        else:
            module_name, callable_name = cache
        callable_object = get_callable(module_name, callable_name)
        if not callable_object:
            continue
        try:
            callable_object(context)
        except Exception as e:
            pass


# no usage (for the moment)
def get_callable(module_name, callable_name):
    try:
        module = importlib.import_module(module_name)
    except ModuleNotFoundError as e:
        return None
    try:
        callable_object = getattr(module, callable_name)
    except AttributeError as e:
        return None
    return callable_object


class IndexBuilder:
    @staticmethod
    def build(dossier):
        index_path = os.path.join(dossier, "index")
        index = list()
        if os.path.isfile(index_path):
            index = IndexParser.parse(index_path)
        registered_targets = {item.filename for item in index}
        unordered_files = IndexBuilder._get_unordered_files(dossier)
        ordered_files = sorted(unordered_files, key=itemgetter(2))
        for item in ordered_files:
            target, base, _ = item
            if target in registered_targets:
                continue
            title, tags = IndexBuilder._get_title_and_tags(base)
            page = dto.PageInfo(target, 0, title, tags)
            index.append(page)
        IndexBuilder._save_index(index, index_path)

    @staticmethod
    def _get_unordered_files(dossier):
        unordered_files = list()
        for root, dirs, files in os.walk(dossier):
            if os.path.samefile(root, dossier):
                folder = ""
            else:
                relpath = os.path.relpath(root, dossier).strip(os.sep)
                cache = relpath.split(os.sep)
                folder = "/".join(cache)
            for base in files:
                if not base.endswith(".exn"):
                    continue
                filename = os.path.join(root, base)
                target = folder + "/" + base if folder else base
                creation_time = os.path.getctime(filename)
                cache = (target, base, creation_time)
                unordered_files.append(cache)
        return unordered_files

    @staticmethod
    def _get_title_and_tags(base):
        base, ext = os.path.splitext(base)
        tags = base.strip("_").split("_")
        title = base.replace("_", " ").capitalize()
        return title, tags

    @staticmethod
    def _save_index(index, path):
        sections = list()
        for item in index:
            cache = list()
            title = item.title if item.title else "- No title -"
            tags = item.tags if item.tags else ("note", )
            cache.append(item.filename)
            cache.append(title)
            tags = ["#"+tag for tag in tags]
            cache.append(" ".join(tags))
            sections.append("\n".join(cache))
        text = "\n\n".join(sections)
        with open(path, "w") as file:
            file.write(text)


def read_style_file(path):
    if not os.path.isfile(path):
        return
    doc = jesth.read(path)
    section = doc.get("")
    body = section.body if section else list()
    style_data = dict()
    if not body:
        return
    for line in body:
        if not line or line.startswith("#"):
            continue
        k, v = jesth.split_key_value(line)
        if not k:
            continue
        style_data[k] = v
    return style_data



import tkinter as tk
import tkutil
from viewable import Viewable
from exn import constant
from exn.view.roll import Roll


class Past(Viewable):
    def __init__(self, front):
        super().__init__()
        self._front = front
        self._roll = None

    def _create_body(self, parent):
        return tk.Toplevel(parent, name="past")

    def _build(self):
        self.body.bind("<Escape>", lambda e: self._on_close())
        self.body.title("")
        self.body.minsize(*constant.WINDOW_MINSIZE)
        tkutil.restore_size(self.body, name="exn_past",
                            default=constant.PAST_WINDOW_SIZE)

        #self.body.resizable(False, False)
        # label title
        frame1 = tk.Frame(self.body)
        frame1.pack(pady=5)
        label = tk.Label(frame1, name="title", text="Past history")
        label.pack()
        # get data and populate roll
        data = self._get_history_data()
        frame2 = tk.Frame(self.body)
        frame2.pack(fill=tk.BOTH, expand=True)
        self._roll = Roll(self._front, data, on_open=self._open_note,
                          on_close=self._on_close)
        self._roll.build_pack(frame2, fill=tk.BOTH, expand=True)
        self._roll.select(0)
        # center the window
        editor = self._front.board.viewer.editor
        tkutil.align(self.body, parent=editor, side="n")
        tkutil.make_modal(self.body)

    def _on_map(self):
        self._roll.body.focus_set()
        pass

    def _on_destroy(self):
        tkutil.save_size(self.body, name="exn_past")

    def _get_history_data(self):
        manager = self._front.manager
        viewer = self._front.board.viewer
        #h = manager.history
        h = manager.dao.get_history()
        data = list()
        for item in h:
            info = viewer.pages.get(item)
            if not info:
                continue
            data.append(info)
        return data

    def _open_note(self, filename):
        self.body.withdraw()
        self._front.open(filename)
        self.body.destroy()

    def _on_close(self):
        self.body.withdraw()
        self.body.destroy()



import os
import os.path
import sys
from exn.manager import Manager
from exonote import constant, split_target
from exn import utils


HELP_TEXT = """
oooooooooooo      ooooooo  ooooo      ooooo      ooo 
`888'     `8       `8888    d8'       `888b.     `8' 
 888                 Y888..8P          8 `88b.    8  
 888oooo8             `8888'           8   `88b.  8  
 888    "            .8PY888.          8     `88b.8  
 888       o        d8'  `888b         8       `888  
o888ooooood8      o888o  o88888o      o8o        `8  
                                                     

This software is the official Exonote Reader
to read a dossier of interactive notes.

Usage:
    exn
    exn <filename>
    exn <option> [<arg> ...]

Options:
    -b, --build                     Build the index file
    -r, --restrict [<filename>]     Open a note with low restriction
    -R, --Restrict [<filename>]     Open a note with high restriction
    -h, --help                      Show help text

Warning:
    Open a note with restriction when you don't trust the author !
    Low restriction: block the execution of embedded programs.
    High restriction: same as low restriction + block executable links

Tip:
    You can replace a filename with its associated page number !


Visit the webpage: https://github.com/pyrustic/exn
"""


def main():
    args = sys.argv[1:]
    if not args:
        open_exonote()
        return
    first_arg = args[0]
    # open target
    if not first_arg.startswith("-"):
        open_exonote(target=first_arg)
        return
    # build index
    if first_arg in ("-b", "--build") and len(args) == 1:
        build_index()
    # open a note with a low or high restriction
    elif first_arg in ("-r", "--restrict", "-R", "--Restrict"):
        restriction = constant.LOW
        if first_arg in ("-R", "--Restrict"):
            restriction = constant.HIGH
        if len(args) == 1:
            open_exonote(restriction=restriction)
        else:
            open_exonote(target=args[1], restriction=restriction)
    # show the help
    else:
        print(HELP_TEXT)


def open_exonote(target=None, restriction=constant.ZERO):
    if target:
        target = update_target(target)
    dossier = os.getcwd()
    manager = Manager(dossier, restriction)
    manager.start(target)



def build_index():
    dossier = os.getcwd()
    utils.IndexBuilder.build(dossier)
    print("Index successfully built !")


def update_target(target):
    target = target.strip("'\"")
    filename, sid = split_target(target)
    try:
        page = int(filename)
    except (ValueError, TypeError) as e:
        pass
    else:
        filename = get_filename_by_page(page)
    if not filename:
        return
    return filename + sid


def get_filename_by_page(page):
    dossier = os.getcwd()
    index_path = os.path.join(dossier, "index")
    index = utils.IndexParser.parse(index_path)
    if not index:
        return
    try:
        info = index[page - 1]
    except IndexError as e:
        return
    return info.filename


if __name__ == "__main__":
    main()



import tkinter as tk
from tkinter import ttk
import tkutil
from viewable import Viewable
from exn import constant
from exn.view.roll import Roll


class Switcher(Viewable):
    def __init__(self, front):
        super().__init__()
        self._front = front
        self._scrolled_text = None
        self._current_index = 0
        self._previous_index = None
        self._history = list()
        self._widgets = list()

    @property
    def front(self):
        return self._front

    def _create_body(self, parent):
        return tk.Toplevel(parent, name="switcher")

    def _build(self):
        self.body.bind("<Tab>", lambda e: "break", add=True)
        self.body.bind("<KeyRelease>", self._on_key_release, add=True)
        self.body.title("")
        self.body.minsize(*constant.WINDOW_MINSIZE)
        tkutil.restore_size(self.body, name="exn_switcher",
                            default=constant.SWITCHER_WINDOW_SIZE)
        # label title
        frame1 = ttk.Frame(self.body)
        frame1.pack()
        label = tk.Label(frame1, name="title", text="Switcher")
        label.pack()
        frame2 = ttk.Frame(self.body)
        frame2.pack(fill=tk.BOTH, expand=True)
        # get history data and populate roll
        self._history = self._get_history_data()
        self._roll = Roll(self._front, self._history,
                          on_close=self._on_close,
                          on_open=self._open_note)
        self._roll.build_pack(frame2, fill=tk.BOTH, expand=True)
        # center window
        editor = self._front.board.viewer.editor
        tkutil.center(self.body, parent=editor)
        tkutil.make_modal(self.body)

    def _on_map(self):
        if not self._history:
            self.body.destroy()
            return
        current_index = 0
        if len(self._history) >= 2:
            current_index = 1
        self._roll.select(current_index)

    def _on_destroy(self):
        tkutil.save_size(self.body, name="exn.view.switcher")

    def _on_key_release(self, e):
        keysym = e.keysym
        if keysym in constant.CONTROL_KEYS:
            info = self._roll.get_selection()
            if info:
                self._open_note(info)
            self.body.destroy()

    def _on_select_prev(self, e):
        return "break"

    def _on_select_next(self, e):
        return "break"

    def _open_note(self, info):
        self._front.open(info.filename)

    def _get_history_data(self):
        manager = self._front.manager
        viewer = self._front.board.viewer
        #h = manager.history
        h = manager.dao.get_history()
        data = list()
        for item in h:
            info = viewer.pages.get(item)
            if not info:
                continue
            data.append(info)
        return data

    def _on_close(self):
        self.body.destroy()



from tkinter import font
from tkinter.ttk import Style as TtkStyle


def get_standard_font_family(name="TkDefaultFont"):
    return font.nametofont(name).actual()["family"]


class Constants:
    FONT_FAMILY = "Helvetica"
    FONT_FAMILY_MONO = "Courier"
    BUTTON_FONT_SIZE = 10
    TEXT_FONT_SIZE = 11
    TOPLEVEL_BACKGROUND = "#F1EEFF"
    TOC_ENTRY_FOREGROUND = "#4A7BF8"
    TOC_ENTRY_ACTIVE_BACKGROUND = "#E2E2FF"
    ROLL_ENTRY_HIGHLIGHT = "#E2DFFF"
    ROLL_ENTRY_ACTIVE_HIGHLIGHT = "#9AA7B8"
    ROLL_ENTRY_BACKGROUND = "#E7E4FF"


class Theme:
    def __init__(self, root):
        self._root = root
        self._ttkstyle = TtkStyle()
        self._update_constants()
        self._set_classic_theme()
        self._set_modern_theme()

    def _update_constants(self):
        Constants.FONT_FAMILY = get_standard_font_family("TkTextFont")
        Constants.FONT_FAMILY_MONO = get_standard_font_family("TkFixedFont")

    def _set_classic_theme(self):
        # Frames
        self._root.option_add("*Frame.background", "white")
        # windows
        self._set_classic_window_style()
        # Text editor
        self._root.option_add("*Text.highlightThickness", 0)
        self._root.option_add("*Text.borderWidth", 0)
        # set classic entry style
        self._set_classic_entry_style()
        # set page number style
        self._set_classic_page_number_style()
        # set finder status
        self._set_classic_finder_status_style()
        # TOC editor
        self._root.option_add("*toc*Text.background", Constants.TOPLEVEL_BACKGROUND)
        self._root.option_add("*toc*Text.highlightThickness", 0)
        self._root.option_add("*toc*Text.borderWidth", 0)
        # Toc info card
        self._set_classic_info_card_style()
        # Goto style
        self._set_classic_goto_style()
        # Roll entry card
        self._set_classic_roll_entry_card_style()
        # Roll nav status style
        self._set_classic_roll_nav_status_style()
        # Search window style
        self._set_classic_search_window_style()
        # About window style
        self._set_classic_about_window_style()

    def _set_modern_theme(self):
        self._ttkstyle.theme_use("classic")
        self._set_modern_frame_style()
        self._set_modern_button_style()
        self._set_modern_clear_button_style()
        self._set_modern_red_button_style()
        self._set_modern_scrollbar_style()
        self._set_modern_window_scrollbar_style()
        self._set_modern_checkbutton_style()
        self._set_modern_checkbutton2_style()
        self._set_modern_page_number_style()
        self._set_modern_mark_button_style()
        self._set_modern_mark_small_button_style()
        self._set_modern_blue_button_style()
        # calc style for the demo exclusively
        self._set_calc_style()

    # ====== CLASSIC THEME =======

    def _set_classic_window_style(self):
        # frames
        self._root.option_add("*Toplevel.background", Constants.TOPLEVEL_BACKGROUND)
        self._root.option_add("*Toplevel*Frame.background", Constants.TOPLEVEL_BACKGROUND)
        # title
        self._root.option_add("*Toplevel*title.foreground", "#505050")
        self._root.option_add("*Toplevel*title.background", Constants.TOPLEVEL_BACKGROUND)
        self._root.option_add("*Toplevel*title.font", (Constants.FONT_FAMILY_MONO,
                                                       Constants.TEXT_FONT_SIZE + 10, "bold"))

    def _set_classic_entry_style(self):
        # entries
        self._root.option_add("*Entry.font", (Constants.FONT_FAMILY_MONO,
                                              Constants.TEXT_FONT_SIZE, "normal"))
        self._root.option_add("*Entry.background", "white")
        self._root.option_add("*Entry.readonlyBackground", "white")
        self._root.option_add("*Entry.disabledBackground", "white")
        self._root.option_add("*Entry.foreground", "#64717A")
        self._root.option_add("*Entry.relief", "flat")
        self._root.option_add("*Entry.highlightThickness", 1)
        self._root.option_add("*Entry.highlightBackground", "#DCDCDC")
        self._root.option_add("*Entry.highlightColor", "#CCCCCC")
        self._root.option_add("*Entry.selectBackground", "#DCE7FF")
        self._root.option_add("*Entry.selectForeground", "#64717A")
        self._root.option_add("*Entry.insertBackground", "#64717A")
        #self._root.option_add("*Entry.padX", 3)
        #self._root.option_add("*Entry.padY", 10)

    def _set_classic_page_number_style(self):
        self._root.option_add("*navframe.page_number.font",
                              (Constants.FONT_FAMILY_MONO,
                               Constants.TEXT_FONT_SIZE , "bold"))
        self._root.option_add("*navframe.page_number.background", "white")
        self._root.option_add("*navframe.page_number.readonlyBackground", "white")
        self._root.option_add("*navframe.page_number.disabledBackground", "white")
        self._root.option_add("*navframe.page_number.foreground", "#64717A")
        self._root.option_add("*navframe.page_number.relief", "flat")
        self._root.option_add("*navframe.page_number.highlightThickness", 1)
        self._root.option_add("*navframe.page_number.highlightBackground", "white")
        self._root.option_add("*navframe.page_number.highlightColor", "#CCCCCC")
        self._root.option_add("*navframe.page_number.selectBackground", "#DCE7FF")
        self._root.option_add("*navframe.page_number.selectForeground", "#64717A")
        self._root.option_add("*navframe.page_number.insertBackground", "#64717A")

    def _set_classic_finder_status_style(self):
        # entries
        self._root.option_add("*finder_status*highlightThickness", 0)

    def _set_classic_info_card_style(self):
        # outer
        self._root.option_add("*info_card.background",
                              Constants.ROLL_ENTRY_BACKGROUND)
        self._root.option_add("*info_card.highlightThickness", 5)
        self._root.option_add("*info_card.highlightBackground",
                              Constants.ROLL_ENTRY_HIGHLIGHT)
        # inner
        self._root.option_add("*info_card*Frame.background",
                              Constants.ROLL_ENTRY_BACKGROUND)
        self._root.option_add("*info_card*Frame.highlightThickness", 0)
        # title field
        self._root.option_add("*info_card*title_field.disabledBackground",
                              Constants.ROLL_ENTRY_BACKGROUND)
        self._root.option_add("*info_card*title_field.disabledForeground", "#2F5F64")
        self._root.option_add("*info_card*title_field.highlightThickness", 0)
        self._root.option_add("*info_card*title_field.borderWidth", 0)
        self._root.option_add("*info_card*title_field.font",
                              (Constants.FONT_FAMILY,
                               Constants.TEXT_FONT_SIZE, "normal"))
        # filename field
        self._root.option_add("*info_card*filename_field.disabledBackground",
                              Constants.ROLL_ENTRY_BACKGROUND)
        self._root.option_add("*info_card*filename_field.disabledForeground", "#575757")
        self._root.option_add("*info_card*filename_field.highlightThickness", 0)
        self._root.option_add("*info_card*filename_field.borderWidth", 0)
        self._root.option_add("*info_card*filename_field.font",
                              (Constants.FONT_FAMILY_MONO,
                               Constants.TEXT_FONT_SIZE - 1, "normal"))
        # label page
        self._root.option_add("*info_card*page_label.background", Constants.ROLL_ENTRY_BACKGROUND)
        self._root.option_add("*info_card*page_label.foreground", "#2F5F64")
        self._root.option_add("*info_card*page_label.font",
                              (Constants.FONT_FAMILY,
                               Constants.TEXT_FONT_SIZE - 1, "bold"))
        # tags field
        self._root.option_add("*info_card*tags_field.disabledBackground",
                              Constants.ROLL_ENTRY_BACKGROUND)
        self._root.option_add("*info_card*tags_field.disabledForeground", "gray")
        self._root.option_add("*info_card*tags_field.highlightThickness", 0)
        self._root.option_add("*info_card*tags_field.borderWidth", 0)
        self._root.option_add("*info_card*tags_field.font",
                              (Constants.FONT_FAMILY,
                               Constants.TEXT_FONT_SIZE - 2, "normal"))

    def _set_classic_roll_entry_card_style(self):
        # outer
        self._root.option_add("*roll*main_frame.Frame.Frame.background",
                              Constants.ROLL_ENTRY_BACKGROUND)
        self._root.option_add("*roll*main_frame.Frame.Frame.highlightThickness", 5)
        self._root.option_add("*roll*main_frame.Frame.Frame.highlightBackground",
                              Constants.ROLL_ENTRY_HIGHLIGHT)
        # inner
        self._root.option_add("*roll*main_frame.Frame.Frame*Frame.background",
                              Constants.ROLL_ENTRY_BACKGROUND)
        self._root.option_add("**roll*main_frame.Frame.Frame*Frame.highlightThickness", 0)
        # title field
        self._root.option_add("*roll*main_frame*title_field.disabledBackground",
                              Constants.ROLL_ENTRY_BACKGROUND)
        self._root.option_add("*roll*main_frame*title_field.disabledForeground", "#2F5F64")
        self._root.option_add("*roll*main_frame*title_field.highlightThickness", 0)
        self._root.option_add("*roll*main_frame*title_field.borderWidth", 0)
        self._root.option_add("*roll*main_frame*title_field.font",
                              (Constants.FONT_FAMILY,
                               Constants.TEXT_FONT_SIZE, "normal"))
        # filename field
        self._root.option_add("*roll*main_frame*filename_field.disabledBackground",
                              Constants.ROLL_ENTRY_BACKGROUND)
        self._root.option_add("*roll*main_frame*filename_field.disabledForeground", "#575757")
        self._root.option_add("*roll*main_frame*filename_field.highlightThickness", 0)
        self._root.option_add("*roll*main_frame*filename_field.borderWidth", 0)
        self._root.option_add("*roll*main_frame*filename_field.font",
                              (Constants.FONT_FAMILY_MONO,
                               Constants.TEXT_FONT_SIZE - 1, "normal"))
        # label page
        self._root.option_add("*roll*main_frame*page_label.background", Constants.ROLL_ENTRY_BACKGROUND)
        self._root.option_add("*roll*main_frame*page_label.foreground", "#2F5F64")
        self._root.option_add("*roll*main_frame*page_label.font",
                              (Constants.FONT_FAMILY,
                               Constants.TEXT_FONT_SIZE - 1, "bold"))
        # tags field
        self._root.option_add("*roll*main_frame*tags_field.disabledBackground",
                              Constants.ROLL_ENTRY_BACKGROUND)
        self._root.option_add("*roll*main_frame*tags_field.disabledForeground", "gray")
        self._root.option_add("*roll*main_frame*tags_field.highlightThickness", 0)
        self._root.option_add("*roll*main_frame*tags_field.borderWidth", 0)
        self._root.option_add("*roll*main_frame*tags_field.font",
                              (Constants.FONT_FAMILY,
                               Constants.TEXT_FONT_SIZE - 2, "normal"))

    def _set_classic_goto_style(self):
        # entry
        self._root.option_add("*goto*Entry.font",
                              (Constants.FONT_FAMILY_MONO,
                               Constants.TEXT_FONT_SIZE, "normal"))
        self._root.option_add("*goto*Entry.background", "#F6F3FF")
        self._root.option_add("*goto*Entry.readonlyBackground", "white")
        self._root.option_add("*goto*Entry.disabledBackground", "white")
        self._root.option_add("*goto*Entry.foreground", "#64717A")
        self._root.option_add("*goto*Entry.relief", "flat")
        self._root.option_add("*goto*Entry.highlightThickness", 1)
        self._root.option_add("*goto*Entry.highlightBackground", "#DCDCDC")
        self._root.option_add("*goto*Entry.highlightColor", "#CCCCCC")
        self._root.option_add("*goto*Entry.selectBackground", "#DCE7FF")
        self._root.option_add("*goto*Entry.selectForeground", "#64717A")
        self._root.option_add("*goto*Entry.insertBackground", "#64717A")
        self._root.option_add("*goto*Entry.padX", 3)
        self._root.option_add("*goto*Entry.padY", 0)
        # description label style
        self._root.option_add("*goto*description_label.background",
                              Constants.TOPLEVEL_BACKGROUND)
        self._root.option_add("*goto*description_label.foreground", "#64717A")
        self._root.option_add("*goto*description_label.foreground", "#3B4851")
        self._root.option_add("*goto*description_label.font",
                              (Constants.FONT_FAMILY,
                               Constants.TEXT_FONT_SIZE, "normal"))
        # status label style
        self._root.option_add("*goto*status_label.background",
                              Constants.TOPLEVEL_BACKGROUND)
        self._root.option_add("*goto*status_label.foreground", "tomato")
        self._root.option_add("*goto*status_label.font",
                              (Constants.FONT_FAMILY, Constants.TEXT_FONT_SIZE, "normal"))

    def _set_classic_roll_nav_status_style(self):
        self._root.option_add("*roll*nav_status.font",
                              (Constants.FONT_FAMILY_MONO,
                               Constants.TEXT_FONT_SIZE, "bold"))
        self._root.option_add("*roll*nav_status.background",
                              Constants.TOPLEVEL_BACKGROUND)
        self._root.option_add("*roll*nav_status.foreground", "#64717A")
        self._root.option_add("*roll*nav_status.activeBackground", "#111519")
        self._root.option_add("*roll*nav_status.activeForeground", "#A098A0")
        self._root.option_add("*roll*nav_status.highlightBackground", "#D0D0D0")
        self._root.option_add("*roll*nav_status.highlightThickness", 0)
        self._root.option_add("*roll*nav_status.highlightColor", "white")
        self._root.option_add("*roll*nav_status.borderWidth", 0)
        self._root.option_add("*roll*nav_status.cursor", "arrow")

    def _set_classic_search_window_style(self):
        # entry
        self._root.option_add("*search*search_field.font",
                              (Constants.FONT_FAMILY_MONO,
                               Constants.TEXT_FONT_SIZE, "normal"))
        self._root.option_add("*search*search_field.background", "#F6F3FF")
        self._root.option_add("*search*search_field.readonlyBackground", "white")
        self._root.option_add("*search*search_field.disabledBackground", "white")
        self._root.option_add("*search*search_field.foreground", "#64717A")
        self._root.option_add("*search*search_field.relief", "flat")
        self._root.option_add("*search*search_field.highlightThickness", 1)
        self._root.option_add("*search*search_field.highlightBackground", "#DCDCDC")
        self._root.option_add("*search*search_field.highlightColor", "#CCCCCC")
        self._root.option_add("*search*search_field.selectBackground", "#DCE7FF")
        self._root.option_add("*search*search_field.selectForeground", "#64717A")
        self._root.option_add("*search*search_field.insertBackground", "#64717A")
        self._root.option_add("*search*search_field.padX", 3)
        self._root.option_add("*search*search_field.padY", 0)

    def _set_classic_about_window_style(self):
        # logo
        self._root.option_add("*about*logo.background", Constants.TOPLEVEL_BACKGROUND)
        self._root.option_add("*about*logo.foreground", "#64717A")
        self._root.option_add("*about*logo.foreground", "#8A9EB0")
        self._root.option_add("*about*logo.foreground", "#F28D71")
        self._root.option_add("*about*logo.font",
                              (Constants.FONT_FAMILY,
                               Constants.TEXT_FONT_SIZE + 42, "bold"))
        # text
        self._root.option_add("*about*Text.font",
                              (Constants.FONT_FAMILY_MONO,
                               Constants.TEXT_FONT_SIZE, "normal"))
        self._root.option_add("*about*Text.background", "#F6F3FF")
        self._root.option_add("*about*Text.background", Constants.TOPLEVEL_BACKGROUND)
        self._root.option_add("*about*Text.readonlyBackground", "white")
        self._root.option_add("*about*Text.disabledBackground", "white")
        self._root.option_add("*about*Text.foreground", "#64717A")
        self._root.option_add("*about*Text.relief", "flat")
        self._root.option_add("*about*Text.highlightThickness", 0)
        self._root.option_add("*about*Text.highlightBackground", "#E7E4FF")
        self._root.option_add("*about*Text.highlightColor", "#CCCCCC")
        self._root.option_add("*about*Text.inactiveSelectBackground", "#DCE7FF")
        self._root.option_add("*about*Text.selectForeground", "#64717A")
        self._root.option_add("*about*Text.insertBackground", "#64717A")
        self._root.option_add("*about*Text.padX", 3)
        self._root.option_add("*about*Text.padY", 0)

    # ====== MODERN THEME =======

    def _set_modern_frame_style(self):
        self._ttkstyle.configure("TFrame", background="white")

    def _set_modern_button_style(self):
        self._ttkstyle.configure("TButton",
                                 font=(Constants.FONT_FAMILY_MONO,
                                       Constants.BUTTON_FONT_SIZE, "normal"),
                                 background="#F2F2F2",
                                 foreground="#838383", relief="flat",
                                 borderwidth=0, padding=(3, 0),
                                 highlightthickness=1,
                                 highlightcolor="#D0D0D0")
        self._ttkstyle.map("TButton", background=[("active", "#E6E6E6")],
                           highlightcolor=[("focus", "gray")])

    def _set_modern_red_button_style(self):
        self._ttkstyle.configure("red.TButton",
                                 font=(Constants.FONT_FAMILY_MONO,
                                       Constants.BUTTON_FONT_SIZE, "normal"),
                                 background="#FFEBFF",
                                 foreground="#F08EAC", relief="flat",
                                 borderwidth=0, padding=(3, 0),
                                 highlightthickness=1,
                                 highlightcolor="#ECAEC8")
        self._ttkstyle.map("red.TButton", background=[("active", "#FFD2EC")],
                           highlightcolor=[("focus", "gray")])

    def _set_modern_blue_button_style(self):
        self._ttkstyle.configure("blue.TButton",
                                 font=(Constants.FONT_FAMILY_MONO,
                                       Constants.BUTTON_FONT_SIZE, "normal"),
                                 background="#E7E4FF",
                                 foreground="#8A9EB0", relief="flat",
                                 borderwidth=0, padding=(3, 0),
                                 highlightthickness=1,
                                 highlightcolor="#CECBE6")
        self._ttkstyle.map("blue.TButton", background=[("active", "#DDDAFF")],
                           highlightcolor=[("focus", "gray")])

    def _set_modern_clear_button_style(self):
        self._ttkstyle.configure("clear.TButton",
                                 font=(Constants.FONT_FAMILY_MONO,
                                       Constants.BUTTON_FONT_SIZE, "normal"),
                                 background="white",
                                 foreground="#838383", relief="flat",
                                 borderwidth=0, padding=0,
                                 highlightcolor="white",
                                 highlightthickness=1)
        self._ttkstyle.map("clear.TButton", background=[("active", "#E6E6E6")],
                           highlightcolor=[("focus", "gainsboro")])
        self._set_modern_clear_button_style2()

    def _set_modern_clear_button_style2(self):
        self._ttkstyle.configure("clear2.TButton",
                                 font=(Constants.FONT_FAMILY_MONO,
                                       Constants.BUTTON_FONT_SIZE, "normal"),
                                 background=Constants.TOPLEVEL_BACKGROUND,
                                 foreground="#838383", relief="flat",
                                 borderwidth=0, padding=0,
                                 highlightcolor=Constants.TOPLEVEL_BACKGROUND,
                                 highlightthickness=1)
        self._ttkstyle.map("clear2.TButton", background=[("active", "#E6E6E6")],
                           highlightcolor=[("focus", "gainsboro")])

    def _set_modern_scrollbar_style(self):
        self._ttkstyle.configure("Vertical.TScrollbar",
                                 relief="flat", borderwidth=0,
                                 troughcolor="white", width=11,
                                 arrowsize=0,
                                 background="#E7E7E7")
        self._ttkstyle.map("Vertical.TScrollbar", background=[("active", "#D1D1D1")],
                           troughcolor=[("active", "#F4F4F4")])

    def _set_modern_window_scrollbar_style(self):
        self._ttkstyle.configure("window.Vertical.TScrollbar",
                                 relief="flat", borderwidth=0,
                                 troughcolor=Constants.TOPLEVEL_BACKGROUND,
                                 width=11, arrowsize=0,
                                 background="#E7E7E7")
        self._ttkstyle.map("window.Vertical.TScrollbar", background=[("active", "#D1D1D1")],
                           troughcolor=[("active", "#F4F4F4")])

    def _set_modern_checkbutton_style(self):
        self._ttkstyle.configure("TCheckbutton",
                                 font=(Constants.FONT_FAMILY_MONO,
                                       Constants.BUTTON_FONT_SIZE, "normal"),
                                 background="white",
                                 foreground="#838383", relief="flat",
                                 borderwidth=0, padding=(5, 0),
                                 highlightcolor="white",
                                 indicatorcolor="gainsboro",
                                 highlightthickness=1)
        self._ttkstyle.map("TCheckbutton", background=[("active", "white")],
                           highlightcolor=[("focus", "gainsboro")],
                           indicatorcolor=[("selected", "gray")])

    def _set_modern_checkbutton2_style(self):
        self._ttkstyle.configure("toplevel.TCheckbutton",
                                 font=(Constants.FONT_FAMILY_MONO,
                                       Constants.BUTTON_FONT_SIZE, "normal"),
                                 background=Constants.TOPLEVEL_BACKGROUND,
                                 foreground="#838383", relief="flat",
                                 borderwidth=0, padding=(5, 0),
                                 highlightcolor=Constants.TOPLEVEL_BACKGROUND,
                                 indicatorcolor="gainsboro",
                                 highlightthickness=1)
        self._ttkstyle.map("toplevel.TCheckbutton", background=[("active", "white")],
                           highlightcolor=[("focus", "gainsboro")],
                           indicatorcolor=[("selected", "gray")])

    def _set_modern_page_number_style(self):
        self._ttkstyle.configure("page_number.TEntry",
                                 font=(Constants.FONT_FAMILY_MONO,
                                       Constants.TEXT_FONT_SIZE, "bold"),
                                 background="white",
                                 foreground="#64717A", relief="flat",
                                 borderwidth=0, padding=0,
                                 highlightcolor="white",
                                 indicatorcolor="gainsboro",
                                 highlightthickness=0)

    def _set_modern_mark_button_style(self):
        self._ttkstyle.configure("mark.TButton",
                                 font=(Constants.FONT_FAMILY_MONO,
                                       Constants.BUTTON_FONT_SIZE, "normal"),
                                 background="#F2F2C0",
                                 foreground="#A8A860", relief="flat",
                                 borderwidth=0, padding=(3, 0),
                                 highlightthickness=1,
                                 highlightcolor="#D0D0D0")
        self._ttkstyle.map("mark.TButton", background=[("active", "#F2F2A5")],
                           highlightcolor=[("focus", "gray")])

    def _set_modern_mark_small_button_style(self):
        self._ttkstyle.configure("mark_small.TButton",
                                 font=(Constants.FONT_FAMILY_MONO,
                                       Constants.BUTTON_FONT_SIZE - 2, "normal"),
                                 background="#F2F2C0",
                                 foreground="#A8A860", relief="flat",
                                 borderwidth=0, padding=(3, 0),
                                 highlightthickness=1,
                                 highlightcolor="#D0D0D0")
        self._ttkstyle.map("mark_small.TButton", background=[("active", "#F2F2A5")],
                           highlightcolor=[("focus", "gray")])
                         
    def _set_calc_style(self):
        # frame
        bg = "#FAFAFA"
        self._root.option_add("*calc.background", bg)
        self._root.option_add("*calc.highlightThickness", 1)
        self._root.option_add("*calc.highlightBackground", "#CCCCCC")
        self._root.option_add("*calc.highlightColor", "#CCCCCC")
        self._root.option_add("*calc*Frame.background", bg)
        # entry
        self._root.option_add("*calc*Entry.font",
                              (Constants.FONT_FAMILY_MONO,
                               Constants.TEXT_FONT_SIZE + 2, "normal"))
        self._root.option_add("*calc*Entry.background", "#F6F3FF")
        self._root.option_add("*calc*Entry.readonlyBackground", "white")
        self._root.option_add("*calc*Entry.disabledBackground", "white")
        self._root.option_add("*calc*Entry.foreground", "#8C8C9A")
        self._root.option_add("*calc*Entry.relief", "flat")
        self._root.option_add("*calc*Entry.highlightThickness", 1)
        self._root.option_add("*calc*Entry.highlightBackground", "#DCDCDC")
        self._root.option_add("*calc*Entry.highlightColor", "#CCCCCC")
        self._root.option_add("*calc*Entry.selectBackground", "#DCE7FF")
        self._root.option_add("*calc*Entry.selectForeground", "#64717A")
        self._root.option_add("*calc*Entry.insertBackground", "#64717A")
        # ttk
        self._ttkstyle.configure("calc.TButton",
                                 font=(Constants.FONT_FAMILY_MONO,
                                       Constants.BUTTON_FONT_SIZE, "bold"),
                                 background="#FAFAFA",
                                 foreground="#9D9DAB", relief="flat",
                                 borderwidth=0, padding=(3, 0),
                                 highlightthickness=1,
                                 highlightcolor="#D0D0D0")
        self._ttkstyle.map("calc.TButton", background=[("active", "#E0E5FF")],
                           highlightcolor=[("focus", "gray")])



import tkinter as tk
from tkinter import ttk
from viewable import Viewable
from exn.view import util
from viewstack import ViewStack
from exn.theme import Constants as ThemeConstants


class Roll(Viewable):
    def __init__(self, front, data, on_open=None,
                 on_close=None, on_leave_focus=None,
                 max_entries=5):
        # target = "all" or "past"
        super().__init__()
        self._front = front
        self._data = data
        self._on_open = on_open
        self._on_close = on_close
        self._on_leave_focus = on_leave_focus
        self._max_entries = max_entries
        self._list_status_strvar = tk.StringVar()
        self._total_lists = 0
        self._main_frame = None
        self._footer_frame = None
        self._nav_frame = None
        self._cache = dict()
        self._current_list_index = 0
        self._current_entry_index = None
        self._entries_frames = list()
        self._viewstack = None

    @property
    def front(self):
        return self._front

    def select(self, entry_index=0):
        if not self.body or not self.body.winfo_exists():
            return
        s = self._viewstack.selection
        if not s:
            return
        self.body.focus_set()
        return s.view.select(entry_index)

    def get_selection(self):
        if not self.body or not self.body.winfo_exists():
            return
        s = self._viewstack.selection
        if not s:
            return
        return s.view.get_selection()

    def open_list(self, list_index):
        if not self.body or not self.body.winfo_exists():
            return
        if list_index != self._current_list_index:
            self._update_roll(list_index)

    def _create_body(self, parent):
        return tk.Frame(parent, name="roll", takefocus=True)

    def _build(self):
        #self.body.bind("<FocusIn>", lambda e: self._on_focus_in())
        self.body.bind("<Up>", lambda e: self._on_press_up())
        self.body.bind("<Down>", lambda e: self._on_press_down())
        self.body.bind("<Left>", lambda e: self._on_press_left())
        self.body.bind("<Right>", lambda e: self._on_press_right())
        self.body.bind("<Return>", lambda e: self._on_open_selection())
        self.body.bind("<Tab>", lambda e: self._on_press_down())
        self._install_main_frame()
        self._install_footer()
        self._update_total_lists_var()
        self._update_roll(0, None)

    def _on_map(self):
        pass

    def _install_main_frame(self):
        # install roll
        self._main_frame = tk.Frame(self.body, name="main_frame", takefocus=False)
        self._main_frame.pack(side=tk.TOP, fill=tk.BOTH, expand=True, pady=5)
        self._main_frame.columnconfigure(0, weight=1)
        self._main_frame.rowconfigure(0, weight=1)
        # init viewstack
        self._viewstack = ViewStack(self._main_frame)

    def _install_footer(self):
        # install footer
        self._footer_frame = tk.Frame(self.body, takefocus=False)
        self._footer_frame.pack(side=tk.BOTTOM, fill=tk.X)
        button_close = ttk.Button(self._footer_frame, style="red.TButton",
                                  name="close", text="Close", command=self._on_close)
        button_close.pack(side=tk.LEFT, padx=3, pady=3)
        self._install_nav_frame()

    def _install_nav_frame(self):
        self._nav_frame = tk.Frame(self._footer_frame, name="nav_frame", takefocus=False)
        self._nav_frame.pack(side=tk.RIGHT, padx=3, pady=3)
        # button next
        next_button = ttk.Button(self._nav_frame, text="Next", command=self._on_click_next)
        next_button.pack(side=tk.RIGHT, padx=(3, 0))
        # button prev
        prev_button = ttk.Button(self._nav_frame, text="Prev", command=self._on_click_prev)
        prev_button.pack(side=tk.RIGHT, padx=(3, 0))
        # status
        nav_status = tk.Entry(self._nav_frame, name="nav_status",
                              justify=tk.RIGHT, width=0, takefocus=False,
                              textvariable=self._list_status_strvar)
        nav_status.pack(side=tk.RIGHT, padx=(3, 10), fill=tk.X, expand=True)

    def _update_roll(self, list_index=0, entry_index=None):
        self._update_total_lists_status()
        if list_index < 0 or list_index >= len(self._data):
            return
        list_name = "list_{}".format(list_index)
        if list_name in self._viewstack.views:
            self._viewstack.lift(list_name)
            view = self._viewstack.selection.view
        else:
            view = EntriesList(self._front, self._data, list_index,
                               self._on_open, self._on_leave_focus,
                               max_entries=5)
            self._viewstack.add(list_name, view)
        if entry_index is not None:
            view.select(entry_index)

    def _on_press_left(self):
        self._on_click_prev()

    def _on_press_right(self):
        self._on_click_next()

    def _on_click_next(self):
        self.body.focus_set()
        self._unhighlight_selection()
        self._current_entry_index = 0
        self._current_list_index += 1
        if self._current_list_index >= self._total_lists:
            self._current_list_index = 0
        self._update_roll(self._current_list_index, 0)
        self._highlight_selection()

    def _on_click_prev(self):
        self.body.focus_set()
        self._unhighlight_selection()
        self._current_entry_index = 0
        self._current_list_index -= 1
        if self._current_list_index < 0:
            self._current_list_index = self._total_lists - 1
        self._update_roll(self._current_list_index, 0)
        self._highlight_selection()

    def _on_press_up(self):
        s = self._viewstack.selection
        if not s:
            return
        s.view.go_up()

    def _on_press_down(self):
        s = self._viewstack.selection
        if not s:
            return
        s.view.go_down()

    def _on_open_selection(self):
        s = self._viewstack.selection
        if not s:
            return
        s.view.open_selection()

    def _update_total_lists_var(self):
        self._current_list_index = 0
        data_size = len(self._data)
        if data_size <= self._max_entries:
            self._total_lists = 1
            return
        x = data_size // self._max_entries
        y = data_size % self._max_entries
        if y != 0:
            x += 1
        self._total_lists = x

    def _update_total_lists_status(self):
        text = "List {} of {}".format(self._current_list_index + 1,
                                      self._total_lists)
        self._list_status_strvar.set(text)

    def _highlight_selection(self):
        s = self._viewstack.selection
        if not s:
            return
        s.view.highlight_selection()

    def _unhighlight_selection(self):
        s = self._viewstack.selection
        if not s:
            return
        s.view.unhighlight_selection()


class EntriesList(Viewable):
    def __init__(self, front, data, list_index, on_open,
                 on_leave_focus, max_entries=5):
        super().__init__()
        self._front = front
        self._data = data
        self._list_index = list_index
        self._on_open = on_open
        self._on_leave_focus = on_leave_focus
        self._max_entries = max_entries
        self._list_status_strvar = tk.StringVar()
        self._total_lists = 0
        self._main_frame = None
        self._footer_frame = None
        self._nav_frame = None
        self._cache = dict()
        self._selected_entry_index = None
        self._entries_frames = list()

    def select(self, entry_index=0):
        if not self.body:
            return
        if entry_index is None:
            self.unhighlight_selection()
            self._selected_entry_index = None
            return
        n = len(self._entries_frames)
        if entry_index < 0:
            entry_index = n - abs(entry_index)
        if entry_index > n - 1:
            return False
        self.unhighlight_selection()
        self._selected_entry_index = entry_index
        self.highlight_selection()
        return True

    def open_selection(self):
        if self._selected_entry_index is None:
            return
        if self._list_index is None:
            return
        i = (self._list_index * self._max_entries) + self._selected_entry_index
        try:
            item = self._data[i]
        except IndexError as e:
            return False
        self._on_click_open(item.filename)
        return True

    def get_selection(self):
        if self._selected_entry_index is None:
            return
        x = self._max_entries * self._list_index
        x += self._selected_entry_index
        info = self._data[x]
        return info

    def highlight_selection(self):
        if self._selected_entry_index is None:
            return
        try:
            entry_frame = self._entries_frames[self._selected_entry_index]
        except IndexError as e:
            return
        color = ThemeConstants.ROLL_ENTRY_ACTIVE_HIGHLIGHT
        entry_frame.config(highlightbackground=color)
        #entry_frame.focus_set()

    def unhighlight_selection(self):
        if self._selected_entry_index is None:
            return
        try:
            entry_frame = self._entries_frames[self._selected_entry_index]
            self.body.focus_lastfor()
        except IndexError as e:
            return
        color = ThemeConstants.ROLL_ENTRY_HIGHLIGHT
        entry_frame.config(highlightbackground=color)

    def go_up(self):
        self._on_press_up()

    def go_down(self):
        self._on_press_down()

    def _build(self):
        data = self._extract_list_data(self._list_index)
        for i, item in enumerate(data):
            title = item.title if item.title else "- No title -"
            tags = ["#" + tag for tag in item.tags]
            tags = " ".join(tags)
            ef = self._install_result_entry(self.body, i, item.filename, item.page,
                                            title, tags)
            self._entries_frames.append(ef)

    def _install_result_entry(self, parent, index, filename, page, title, tags):
        frame = tk.Frame(parent, takefocus=False)
        frame.pack(padx=10, pady=5, fill=tk.X)
        # left frame
        left_frame = tk.Frame(frame, takefocus=False)
        left_frame.pack(side=tk.LEFT, fill=tk.X, expand=True,
                        padx=5, pady=5)
        left_frame.bind("<Enter>", lambda e: self.select(index))
        left_frame.bind("<Button-1>",
                        lambda e: self._on_click_open(filename))
        # right frame
        right_frame = tk.Frame(frame, takefocus=False)
        right_frame.pack(side=tk.RIGHT, fill=tk.Y, padx=(0, 5), pady=5)
        right_frame.bind("<Enter>", lambda e: self.select(index))
        right_frame.bind("<Button-1>",
                         lambda e: self._on_click_open(filename))
        # title entry
        title_entry = tk.Entry(left_frame, name="title_field",
                               takefocus=False, cursor="arrow")
        title_entry.pack(fill=tk.X, pady=(0, 2))
        title_entry.insert(0, title)
        title_entry.config(state="disabled")
        title_entry.bind("<Enter>", lambda e: self.select(index))
        title_entry.bind("<Button-1>", lambda e: self._on_click_open(filename))
        command = lambda e: util.update_clipboard(title, self.body)
        title_entry.bind("<Button-3>", command)
        #title_entry.bind("<FocusIn>", lambda e: self.select(index))
        # filename entry and page number
        filename_frame = tk.Frame(left_frame)
        filename_frame.pack(fill=tk.X, pady=(0, 2))
        text = "{}.".format(page)
        page_label = tk.Label(filename_frame, padx=0, name="page_label", text=text)
        page_label.pack(side=tk.LEFT, pady=0)
        filename_entry = tk.Entry(filename_frame, name="filename_field",
                                  takefocus=False, cursor="arrow")
        filename_entry.pack(side=tk.LEFT, fill=tk.X, expand=True, padx=(2, 0))
        filename_entry.insert(0, filename)
        filename_entry.config(state="disabled")
        filename_entry.bind("<Enter>", lambda e: self.select(index))
        filename_entry.bind("<Button-1>", lambda e: self._on_click_open(filename))
        command = lambda e: util.update_clipboard(filename, self.body)
        filename_entry.bind("<Button-3>", command)
        # tags entry
        tags_entry = tk.Entry(left_frame, name="tags_field",
                              takefocus=False, cursor="arrow")
        tags_entry.pack(fill=tk.X)
        tags_entry.insert(0, tags)
        tags_entry.config(state="disabled")
        tags_entry.bind("<Enter>", lambda e: self.select(index))
        tags_entry.bind("<Button-1>", lambda e: self._on_click_open(filename))
        command = lambda e: util.update_clipboard(tags, self.body)
        tags_entry.bind("<Button-3>", command)
        # mark button
        mark_button = ttk.Button(right_frame, text="Mark", takefocus=False)
        mark_button.pack(side=tk.TOP, pady=(0, 3))
        command = lambda: self._on_click_mark(mark_button, filename)
        mark_button.config(command=command)
        if self._front.manager.dao.is_bookmarked(filename):
            util.set_marked_style(mark_button, small=False)
        # copy button
        text = "{}\n{}. {}\n{}".format(title, page, filename, tags)
        command = lambda: util.update_clipboard(text, self.body)
        copy_button = ttk.Button(right_frame, text="Copy",
                                command=command, takefocus=False)
        copy_button.pack(side=tk.BOTTOM, pady=0)

        return frame

    def _extract_list_data(self, list_index):
        list_index = list_index if list_index == 0 else self._max_entries * list_index
        index_begin = list_index
        index_end = index_begin + self._max_entries
        if not self._data:
            return list()
        return self._data[index_begin:index_end]

    def _update_total_lists_var(self):
        self._list_index = 0
        data_size = len(self._data)
        if data_size <= self._max_entries:
            self._total_lists = 1
            return
        x = data_size // self._max_entries
        y = data_size % self._max_entries
        if y != 0:
            x += 1
        self._total_lists = x

    def _on_press_up(self):
        if self._selected_entry_index is None:
            return
        self.unhighlight_selection()
        self._selected_entry_index -= 1
        if self._selected_entry_index < 0:
            if self._on_leave_focus and self._on_leave_focus():
                self._selected_entry_index = None
                return
            n = len(self._entries_frames)
            self._selected_entry_index = n - 1
        self.highlight_selection()

    def _on_press_down(self):
        if self._selected_entry_index is None:
            return
        self.unhighlight_selection()
        self._selected_entry_index += 1
        n = len(self._entries_frames)
        if self._selected_entry_index >= n:
            if self._on_leave_focus and self._on_leave_focus():
                self._selected_entry_index = None
                return
            self._selected_entry_index = 0
        self.highlight_selection()

    def _on_open_selection(self):
        if not self._on_open:
            return
        info = self.get_selection()
        self._on_open(info.filename)

    def _on_click_open(self, filename):
        if not self._on_open:
            return
        self._on_open(filename)

    def _on_click_mark(self, button, filename):
        util.on_click_mark_button(button, filename,
                                  self._front.manager.dao,
                                  small=False)



import tkinter as tk
import webbrowser
from tkinter import ttk
import tkutil
from viewable import Viewable

TEXT = """\
This software is the official Exonote Reader to read a dossier of interactive notes.

Visit website for documentation and updates: https://github.com/pyrustic/exn\
"""


class About(Viewable):
    def __init__(self, front):
        super().__init__()
        self._front = front

    def _create_body(self, parent):
        return tk.Toplevel(parent, name="about")

    def _build(self):
        self.body.bind("<Escape>", lambda e: self._on_close())
        self.body.title("")
        #self.body.minsize(*constant.WINDOW_MINSIZE)
        #tkutil.restore_size(self.body, name="exn_info",
        #                    default=constant.INFO_WINDOW_SIZE)
        # labels
        frame1 = tk.Frame(self.body)
        frame1.pack(fill=tk.BOTH, expand=True, padx=3, pady=5)
        frame2 = tk.Frame(self.body)
        frame2.pack(fill=tk.X, padx=3, pady=(10, 3))
        # logo
        logo_label = tk.Label(frame1, name="logo", text="EXN")
        logo_label.pack()
        # text
        text = tk.Text(frame1, name="title_field", wrap="word",
                       width=40, height=8,
                       takefocus=False,
                       cursor="arrow")
        text.pack(fill=tk.BOTH, expand=True, pady=5)
        text.insert("1.0", TEXT)
        text.config(state="disabled")
        # footer
        close_button = ttk.Button(frame2, name="close", text="Close",
                                  style="red.TButton",
                                  command=self._on_close)
        close_button.pack(side=tk.LEFT)
        visit_website_button = ttk.Button(frame2, name="website_button",
                                          style="blue.TButton", text="Visit Website",
                                          command=self._on_visit_website)
        visit_website_button.pack(side=tk.RIGHT)
        # center the window
        editor = self._front.board.viewer.editor
        tkutil.align(self.body, parent=editor, side="n")
        tkutil.make_modal(self.body)

    def _on_map(self):
        pass

    def _on_visit_website(self):
        url = "https://github.com/pyrustic/exn"
        command = lambda: webbrowser.open(url, new=0)
        self.body.after(1, command)

    def _on_close(self):
        self.body.withdraw()
        self.body.destroy()



def on_click_mark_button(button, filename, dao, small=True):
    if dao.is_bookmarked(filename):
        dao.unset_bookmark(filename)
        unset_marked_style(button, small)
    else:
        dao.set_bookmark(filename)
        set_marked_style(button, small)


def set_marked_style(button, small=True):
    style_name = "mark.TButton"
    if small:
        style_name = "small_mark.TButton"
    button.config(style=style_name)


def unset_marked_style(button, small=True):
    style_name = "TButton"
    if small:
        style_name = "small.TButton"
    button.config(style=style_name)


def update_clipboard(text, widget):
    widget.clipboard_clear()
    widget.clipboard_append(text)



import tkinter as tk
from tkinter import ttk
import tkutil
from viewable import Viewable
from exn import constant
from tkinter.font import Font
from exonote.editor import Editor
from exn.theme import Constants as theme_const


class Toc(Viewable):
    def __init__(self, front):
        super().__init__()
        self._front = front
        self._editor = None
        self._sections = list()
        self._sid = None
        self._index = -1

    def _create_body(self, parent):
        return tk.Toplevel(parent, name="toc")

    def _build(self):
        self.body.title("")
        self.body.minsize(*constant.WINDOW_MINSIZE)
        self.body.bind("<Escape>", lambda e: self.body.destroy())
        tkutil.restore_size(self.body, name="exn_toc",
                            default=constant.TOC_WINDOW_SIZE)
        #self.body.geometry(constant.DIALOG_GEOMETRY)
        #self.body.resizable(False, False)
        # frame 1
        frame1 = tk.Frame(self.body)
        frame1.pack(pady=5)
        # frame 2
        frame2 = tk.Frame(self.body)
        frame2.pack(fill=tk.BOTH, expand=True)
        # frame 3
        frame3 = tk.Frame(self.body)
        frame3.pack(fill=tk.X)
        # label Toc
        label = tk.Label(frame1, name="title",
                         text="Table of contents")
        label.pack()
        # install scrolled text
        self._editor = Editor(frame2, wrap="word", cursor="arrow", takefocus=True,
                              scrollbar_style="window.Vertical.TScrollbar")
        self._editor.pack(fill=tk.BOTH, expand=True)
        self._editor.config(padx=10, pady=20)
        self._editor.bind("<Left>", lambda e: self._on_press_up())
        self._editor.bind("<Right>", lambda e: self._on_press_down())
        self._editor.bind("<Up>", lambda e: self._on_press_up())
        self._editor.bind("<Down>", lambda e: self._on_press_down())
        self._editor.bind("<Return>", lambda e: self._on_open())
        # button close
        button = ttk.Button(frame3, name="close", style="red.TButton", text="Close",
                           command=self.body.destroy)
        button.pack(side=tk.LEFT, padx=3, pady=3)

        # center the window
        editor = self._front.board.viewer.editor
        tkutil.align(self.body, parent=editor, side="nw",
                     correction=(10, 10))
        tkutil.make_modal(self.body)

    def _on_map(self):
        self._define_tag()
        self._bind_handlers_to_tag()
        sections = self._get_sections()
        self._sections = sections
        self._n = len(sections)
        for section in sections:
            sid = section.sid
            indent = "  " * (section.level)
            text = indent + section.text + "\n"
            self._editor.insert(tk.INSERT, text, ("link", sid))
            self._editor.insert(tk.INSERT, "\n")
            command = lambda e, sid=section.sid: self._on_click_link(sid)
            self._editor.tag_bind(section.sid, "<ButtonPress-1>",
                                  command, True)
            command = lambda e, sid=sid: self._on_enter(sid)
            self._editor.tag_bind(sid, "<Enter>", command, True)
            command = lambda e, sid=sid: self._on_leave(sid)
            self._editor.tag_bind(sid, "<Leave>", command, True)
        height = 3
        if sections:
            n = (len(sections) * 2) - 1
            height = 10
            if n < height:
                height = n
        self._editor.config(height=height, state="disabled")
        self._editor.focus_set()

    def _on_remap(self):
        pass

    def _on_destroy(self):
        tkutil.save_size(self.body, name="exn_toc")

    def _define_tag(self):
        style = self._front.manager.viewer_style
        # link
        font = Font(family=style.font_family,
                    size=style.font_size)
        foreground = theme_const.TOC_ENTRY_FOREGROUND
        spacing1 = spacing3 = 5
        spacing2 = spacing1 + spacing3
        self._editor.tag_configure("link", font=font, spacing1=spacing1,
                                   spacing2=spacing2, spacing3=spacing3,
                                   foreground=foreground)

    def _bind_handlers_to_tag(self):
        on_enter = lambda event, editor=self._editor: editor.config(cursor="hand1")
        on_leave = lambda event, editor=self._editor: editor.config(cursor="")
        # bind hand icon to codeblock (enter vs leave)
        self._editor.tag_bind("link", "<Enter>", on_enter, True)
        #self._editor.tag_bind("link", "<Leave>", on_leave, True)

    def _on_enter(self, sid):
        if self._sid:
            self._unhighlight(self._sid)
        for i, section in enumerate(self._sections):
            if section.sid == sid:
                self._index = i
        self._sid = sid
        self._highlight(sid)

    def _on_leave(self, sid):
        return
        #self._sid = sid
        #self._unhighlight(sid)

    def _highlight(self, sid):
        background = theme_const.TOC_ENTRY_ACTIVE_BACKGROUND
        self._editor.tag_configure(sid, background=background)

    def _unhighlight(self, sid):
        background = theme_const.TOPLEVEL_BACKGROUND
        self._editor.tag_configure(sid, background=background)

    def _get_sections(self):
        sections = list()
        viewer = self._front.board.viewer
        if not viewer:
            return sections
        for sid in viewer.sids:
            info = viewer.get_heading(sid, include_text=True)
            sections.append(info)
        return sections

    def _on_click_link(self, sid):
        self._front.board.viewer.goto(sid)
        self.body.destroy()

    def _on_press_up(self):
        if self._sid:
            self._unhighlight(self._sid)
        self._index -= 1
        n = len(self._sections)
        if self._index < 0:
            self._index = n - 1
        section = self._sections[self._index]
        sid = section.sid
        self._sid = sid
        self._highlight(sid)
        return "break"

    def _on_press_down(self):
        if self._sid:
            self._unhighlight(self._sid)
        self._index += 1
        n = len(self._sections)
        if self._index >= n:
            self._index = 0
        section = self._sections[self._index]
        sid = section.sid
        self._sid = sid
        self._highlight(sid)
        return "break"

    def _on_open(self):
        if not self._sid:
            return
        self._on_click_link(self._sid)



import tkinter as tk
from tkinter import ttk
from viewable import Viewable


class Footer(Viewable):
    def __init__(self, front, on_leave_focus=None):
        super().__init__()
        self._front = front
        self._on_leave_focus = on_leave_focus
        self._page_number = 0
        self._total_pages = 0
        self._finder = None
        self._page_status_strvar = tk.StringVar()
        self._page_status_entry = None
        self._right_frame = None

    @property
    def finder(self):
        return self._finder

    def update_page_status(self, filename):
        viewer = self._front.board.viewer
        pages = viewer.pages
        page_info = pages.get(filename)
        if not page_info:
            return
        self._total_pages = len(pages)
        self._page_number = page_info.page
        text = "Page {} of {}".format(self._page_number,
                                      self._total_pages)
        self._page_status_strvar.set(text)
        #n = len(text)
        #self._page_status_entry.config(width=n)

    def enable_finder(self):
        if not self._finder:
            return
        self._finder.enable()

    def open_prev_page(self):
        self._on_click_prev_page()

    def open_next_page(self):
        self._on_click_next_page()

    def _create_body(self, parent):
        return ttk.Frame(parent, name="footer")

    def _build(self):
        self._install_left_frame()
        self._install_right_frame()

    def _on_map(self):
        pass

    def _install_left_frame(self):
        # left frame
        left_frame = ttk.Frame(self.body)
        left_frame.pack(side=tk.LEFT, fill=tk.X, expand=True, padx=3, pady=3)
        self._finder = Finder(self._front, on_enable=self._hide_right_frame,
                              on_disable=self._unhide_right_frame,
                              on_leave_focus=self._on_leave_focus)
        self._finder.build_pack(left_frame, side=tk.LEFT, fill=tk.X, expand=True)

    def _install_right_frame(self):
        # right frame
        self._right_frame = ttk.Frame(self.body, name="navframe")
        self._right_frame.pack(side=tk.RIGHT, pady=3,
                               fill=tk.X, expand=True)
        # button next
        next_button = ttk.Button(self._right_frame, text="Next",
                                takefocus=False,
                                command=self._on_click_next_page)
        next_button.pack(side=tk.RIGHT, padx=(0, 3))
        # button previous
        prev_button = ttk.Button(self._right_frame, text="Prev",
                                takefocus=False,
                                command=self._on_click_prev_page)
        prev_button.pack(side=tk.RIGHT, padx=(0, 3))
        # page number
        self._page_status_entry = tk.Entry(self._right_frame,
                                           name="page_number",
                                           justify=tk.RIGHT,
                                           width=0,
                                           takefocus=False,
                                           textvariable=self._page_status_strvar)
        self._page_status_entry.pack(side=tk.RIGHT, padx=(0, 15), fill=tk.X,
                                     expand=True)
        self._page_status_entry.bind("<FocusIn>",
                                     lambda e: self._on_page_status_entry_focus())

    def _on_click_next_page(self):
        if self._page_number == self._total_pages:
            next_page_number = 1
        else:
            next_page_number = self._page_number + 1
        self._page_number = next_page_number
        pages = self._front.board.viewer.pages
        page_info = list(pages.values())[next_page_number - 1]
        self._front.open(page_info.filename)
        #self._update_page_status()

    def _on_click_prev_page(self):
        if self._page_number == 0:
            next_page_number = self._total_pages
        else:
            next_page_number = self._page_number - 1
        self._page_number = next_page_number
        pages = self._front.board.viewer.pages
        page_info = list(pages.values())[next_page_number - 1]
        self._front.open(page_info.filename)
        #self._update_page_status()

    def _on_page_status_entry_focus(self):
        if self._on_leave_focus:
            self._on_leave_focus()

    def _hide_right_frame(self):
        if self._right_frame:
            self._right_frame.pack_forget()

    def _unhide_right_frame(self):
        if self._right_frame:
            self._right_frame.pack(side=tk.RIGHT, pady=3,
                                   fill=tk.X, expand=True)


class Finder(Viewable):
    def __init__(self, front, on_enable=None, on_disable=None,
                 on_leave_focus=None):
        super().__init__()
        self._front = front
        self._on_enable = on_enable
        self._on_disable = on_disable
        self._on_leave_focus = on_leave_focus
        self._input_strvar = tk.StringVar()
        self._case_intvar = tk.IntVar()
        self._regex_intvar = tk.IntVar()
        self._status_strvar = tk.StringVar()
        self._finder_entry = None
        self._options_frame = None
        self._nav_frame = None
        self._status_frame = None
        self._cancel_button = None
        self._highlights = list()
        self._highlighted = str()
        self._active_highlight_position = -1

    @property
    def finder_entry(self):
        return self._finder_entry

    def enable(self):
        if self._on_enable:
            self._on_enable()
        keep_input = True
        state = str(self._finder_entry.cget("state"))
        if state == "disabled":
            keep_input = False
        self._clear_finder(keep_input)
        self._options_frame.pack(side=tk.LEFT)
        self._cancel_button.pack(side=tk.RIGHT)
        self._finder_entry.config(cursor="xterm")

    def disable(self):
        if self._on_disable:
            self._on_disable()
        self._clear_finder_state()
        self._input_strvar.set("Find in page...")
        self._finder_entry.config(state="disabled", cursor="hand1")
        self._options_frame.pack_forget()
        self._nav_frame.pack_forget()
        self._status_frame.pack_forget()
        self._cancel_button.pack_forget()
        if self._on_leave_focus:
            self._on_leave_focus()

    def clear_finder(self, keep_input=False):
        self._clear_finder(keep_input)

    def _create_body(self, parent):
        return ttk.Frame(parent)

    def _build(self):
        # find entry
        self._finder_entry = tk.Entry(self.body, name="finder", takefocus=True,
                                      textvariable=self._input_strvar)
        self._finder_entry.pack(side=tk.LEFT)
        #self._finder_entry.bind("<FocusIn>",
        #                        lambda e: self._on_finder_focus(), True)
        self._finder_entry.bind("<Return>",
                                lambda e: self._on_find(), True)
        self._finder_entry.bind("<Up>", lambda e: self._on_click_prev_match())
        self._finder_entry.bind("<Down>", lambda e: self._on_click_next_match())
        self._finder_entry.bind("<Escape>", lambda e: self.disable(), True)
        self._finder_entry.bind("<FocusIn>", lambda e: self.enable(), True)
        self._finder_entry.bind("<Button-1>", lambda e: self.enable(), True)

        # finder controls
        self._options_frame = ttk.Frame(self.body)
        self._options_frame.pack(side=tk.LEFT)
        # button clear
        button_clear = ttk.Button(self._options_frame, style="clear.TButton",
                                  text="X", command=self._clear_finder)
        button_clear.pack(side=tk.LEFT, padx=3)
        # case checkbutton
        case_checkbutton = ttk.Checkbutton(self._options_frame,
                                          text="mc", variable=self._case_intvar,
                                          onvalue=1, offvalue=0)
        case_checkbutton.pack(side=tk.LEFT, padx=3, fill=tk.Y)
        # regex checkbutton
        regex_checkbutton = ttk.Checkbutton(self._options_frame,
                                           text="re", variable=self._regex_intvar,
                                           onvalue=1, offvalue=0)
        regex_checkbutton.pack(side=tk.LEFT, padx=(0, 3), fill=tk.Y)

        self._nav_frame = ttk.Frame(self.body)
        self._nav_frame.pack(side=tk.LEFT)
        # left arrow button
        left_arrow_button = ttk.Button(self._nav_frame, text="<",
                                      command=self._on_click_prev_match)
        left_arrow_button.pack(side=tk.LEFT, padx=(3, 0))
        # right arrow button
        right_arrow_button = ttk.Button(self._nav_frame, text=">",
                                       command=self._on_click_next_match)
        right_arrow_button.pack(side=tk.LEFT, padx=(3, 0))
        # status
        self._status_frame = ttk.Frame(self.body)
        self._status_frame.pack(side=tk.LEFT, fill=tk.X, expand=True)
        self._status_entry = tk.Entry(self._status_frame, name="finder_status",
                                      width=0, takefocus=False,
                                      state="readonly", cursor="arrow",
                                      textvariable=self._status_strvar)
        self._status_entry.pack(side=tk.LEFT, fill=tk.X, expand=True, padx=3)
        # button cancel
        self._cancel_button = ttk.Button(self.body, name="close", style="red.TButton",
                                         text="Cancel",
                                        command=self.disable)
        self._cancel_button.pack(side=tk.RIGHT)

    def _on_map(self):
        self.disable()

    def _on_find(self):
        if self._finder_entry.cget("state") == "disabled":
            return
        word = self._input_strvar.get()
        if not word or word.isspace():
            self._clear_finder()
            return
        if word == self._highlighted:
            self._on_click_next_match()
            return
        self._clear_finder(keep_input=True)
        self._nav_frame.pack(side=tk.LEFT)
        self._status_frame.pack(side=tk.LEFT, fill=tk.X, expand=True)
        self._highlighted = word
        index = "1.0"
        nocase = self._case_intvar.get()
        nocase = False if nocase == 1 else True
        use_regex = self._regex_intvar.get()
        use_regex = True if use_regex == 1 else False
        editor = self._front.board.editor
        # search
        while True:
            try:
                index = editor.search(word, index, stopindex=tk.END,
                                            nocase=nocase, regexp=use_regex)
            except Exception as e:
                return
            if not index:
                break
            index2 = "{}+{}c".format(index, len(word))
            editor.tag_add("highlight", index, index2)
            self._highlights.append((index, index2))
            index = index2
        if self._highlights:
            if len(self._highlights) == 1:
                self._nav_frame.pack_forget()
            self._on_click_next_match()
        else:
            self._clear_finder_state()
            self._clear_finder_ui(keep_input=True)
            self._status_strvar.set("No matches !")
            self._status_frame.pack(side=tk.LEFT, fill=tk.X, expand=True)

    def _on_click_prev_match(self):
        matches_count = len(self._highlights)
        if not matches_count:
            return
        if self._active_highlight_position == 0:
            next_position = matches_count - 1
        else:
            next_position = self._active_highlight_position - 1
        self._active_highlight_position = next_position
        if matches_count == 1:
            status = "1 match"
        else:
            status = "{} of {} matches".format(next_position+1, matches_count)
        self._status_strvar.set(status)
        x, y = self._highlights[next_position]
        editor = self._front.board.editor
        editor.tag_remove("active_highlight", "1.0", tk.END)
        editor.tag_add("active_highlight", x, y)
        editor.see(x)

    def _on_click_next_match(self):
        matches_count = len(self._highlights)
        if not matches_count:
            return
        if self._active_highlight_position == (matches_count - 1):
            next_position = 0
        else:
            next_position = self._active_highlight_position + 1
        self._active_highlight_position = next_position
        if matches_count == 1:
            status = "1 match"
        else:
            status = "{} of {} matches".format(next_position+1, matches_count)
        self._status_strvar.set(status)
        x, y = self._highlights[next_position]
        editor = self._front.board.editor
        editor.tag_remove("active_highlight", "1.0", tk.END)
        editor.tag_add("active_highlight", x, y)
        editor.see(x)

    def _clear_finder(self, keep_input=False):
        self._clear_finder_state()
        self._clear_finder_ui(keep_input)

    def _clear_finder_ui(self, keep_input=False):
        self._finder_entry.config(state="normal")
        if not keep_input:
            self._input_strvar.set("")
        self._nav_frame.pack_forget()
        self._status_frame.pack_forget()
        self._finder_entry.focus_set()

    def _clear_finder_state(self):
        self._highlights = list()
        self._highlighted = str()
        self._active_highlight_position = -1
        self._status_strvar.set("")
        editor = self._front.board.editor
        editor.tag_remove("active_highlight", "1.0", tk.END)
        editor.tag_remove("highlight", "1.0", tk.END)



import tkinter as tk
from tkinter import ttk
import tkutil
from viewable import Viewable


class Goto(Viewable):
    def __init__(self, front):
        super().__init__()
        self._front = front
        self._input_strvar = tk.StringVar()
        self._status_strvar = tk.StringVar()

    def _create_body(self, parent):
        return tk.Toplevel(parent, name="goto")

    def _build(self):
        self.body.bind("<Escape>", lambda e: self._on_close())
        self.body.title("")
        #self.body.minsize(*constant.WINDOW_MINSIZE)
        #tkutil.restore_size(self.body, name="exn_info",
        #                    default=constant.INFO_WINDOW_SIZE)
        # labels
        frame1 = tk.Frame(self.body)
        frame1.pack(pady=5)
        frame2 = tk.Frame(self.body)
        frame2.pack(fill=tk.BOTH, expand=True, padx=3, pady=(5, 0))
        frame3 = tk.Frame(self.body)
        frame3.pack(fill=tk.BOTH, expand=True, padx=3, pady=(3, 5))
        frame4 = tk.Frame(self.body)
        frame4.pack(fill=tk.BOTH, expand=True, padx=3, pady=5)
        frame5 = tk.Frame(self.body)
        frame5.pack(fill=tk.X, padx=3, pady=(10, 3))
        # Info
        info_label = tk.Label(frame1, name="title", text="Goto")
        info_label.pack()
        # description label
        text = "Filename, anchor, or page number"
        description_label = tk.Label(frame2, name="description_label",
                                     text=text, anchor="w")
        description_label.pack(fill=tk.X)
        #
        # entry field
        field = tk.Entry(frame3, name="field", width=30,
                         textvariable=self._input_strvar)
        field.pack(side=tk.LEFT, fill=tk.X, expand=True, padx=(0, 3))
        command = lambda e: self._on_click_goto()
        field.bind("<Return>", command, True)
        field.focus_set()
        # clear button
        command = lambda: self._input_strvar.set("")
        clear_button = ttk.Button(frame3, text="X", style="clear2.TButton",
                                  command=command)
        clear_button.pack(side=tk.RIGHT)
        # status label
        status_label = tk.Label(frame4, name="status_label",
                                textvariable=self._status_strvar)
        status_label.pack(fill=tk.X)
        # footer
        close_button = ttk.Button(frame5, name="close", text="Close",
                                  style="red.TButton", command=self._on_close)
        close_button.pack(side=tk.LEFT)
        goto_button = ttk.Button(frame5, name="goto_button",
                                 text="Goto !", command=self._on_click_goto)
        goto_button.pack(side=tk.RIGHT)
        # center the window
        editor = self._front.board.viewer.editor
        tkutil.align(self.body, parent=editor, side="n")
        tkutil.make_modal(self.body)

    def _on_map(self):
        pass

    def _get_data(self):
        return "title", "filename", "#tag1 #tag2"

    def _on_click_goto(self):
        val = self._input_strvar.get().strip()
        if not val:
            return
        try:
            page = int(val)
        except ValueError as e:
            pass
        else:
            self._open_page(page)
            return
        viewer = self._front.board.viewer
        failure = False
        is_anchor = False
        if val.startswith("<") and val.endswith(">"):
            is_anchor = True
            if not viewer.goto(val):
                failure = True
        else:
            if not self._front.open(val):
                failure = True
        if failure:
            self._status_strvar.set("Failed to process the request !")
        else:
            if is_anchor:
                self._status_strvar.set("Anchor found !")
            else:
                self._status_strvar.set("Page found !")
            self._on_close()

    def _open_page(self, x):
        x -= 1
        pages = self._front.board.viewer.pages
        match = None
        for i, item in enumerate(pages.values()):
            if i == x:
                match = item
                break
        if not match:
            self._status_strvar.set("Page not found !")
            return
        if not self._front.open(match.filename):
            self._status_strvar.set("Failed to open the page")
            return
        self._status_strvar.set("Page found !")
        self._on_close()

    def _on_close(self):
        self.body.withdraw()
        self.body.destroy()



import os
import os.path
import jesth
import atexit


class Dao:
    def __init__(self, dossier):
        self._dossier = dossier
        self._index_filename = None
        self._history_filename = None
        self._bookmarks_filename = None
        self._blocklist_filename = None
        self._history_cache = list()
        self._bookmarks_cache = list()
        self._blocklist = list()
        self._setup()

    def get_history(self):
        return self._history_cache

    def update_history(self, entry):
        history = self.get_history()
        for i, item in enumerate(history):
            if item == entry:
                del history[i]
                break
        history.insert(0, entry)

    def get_bookmarks(self):
        return self._bookmarks_cache

    def set_bookmark(self, path):
        bookmarks = self.get_bookmarks()
        if path not in bookmarks:
            bookmarks.insert(0, path)

    def unset_bookmark(self, path):
        bookmarks = self.get_bookmarks()
        for i, entry in enumerate(bookmarks):
            if entry == path:
                del bookmarks[i]
                break

    def is_bookmarked(self, path):
        bookmarks = self.get_bookmarks()
        for item in bookmarks:
            if item == path:
                return True
        return False

    def get_blocklist(self):
        return self._blocklist

    def _setup(self):
        self._index_filename = os.path.join(self._dossier, "index")
        self._history_filename = os.path.join(self._dossier, ".exn", "history")
        self._bookmarks_filename = os.path.join(self._dossier, ".exn", "bookmarks")
        self._blocklist_filename = os.path.join(self._dossier, ".exn", "blocklist")
        self._load_data()
        atexit.register(self._on_exit)

    def _load_data(self):
        if os.path.exists(self._history_filename):
            doc = jesth.read(self._history_filename)
            cache = doc.get("")
            self._history_cache = cache.body if cache else list()
        if os.path.exists(self._bookmarks_filename):
            doc = jesth.read(self._bookmarks_filename)
            cache = doc.get("")
            self._bookmarks_cache = cache.body if cache else list()
        if os.path.exists(self._blocklist_filename):
            doc = jesth.read(self._blocklist_filename)
            cache = doc.get("")
            self._blocklist = cache.body if cache else list()

    def _save_state(self):
        self._create_dotexn_folder()
        items = [(self._history_filename, self.get_history),
                 (self._bookmarks_filename, self.get_bookmarks)]
        for filename, get_data in items:
            dirname = os.path.dirname(filename)
            if not os.path.isdir(dirname):
                continue
            doc = jesth.Document()
            doc.append("", body=get_data())
            jesth.write(doc, filename)
            
    def _create_dotexn_folder(self):
        dotexn_folder = os.path.join(self._dossier, ".exn")
        if os.path.isdir(dotexn_folder):
            return
        try:
            os.makedirs(dotexn_folder)
        except Exception as e:
            print("Failed to create '.exn' folder")

    def _on_exit(self):
        atexit.unregister(self._on_exit)
        self._save_state()



CONTROL_KEYS = {"Control_L", "Control_R"}
PAST_WINDOW_SIZE = (400, 600)
SWITCHER_WINDOW_SIZE = (400, 600)
TOC_WINDOW_SIZE = (400, 550)
INFO_WINDOW_SIZE = (300, 300)
SEARCH_WINDOW_SIZE = (400, 650)
WINDOW_MINSIZE = (350, 350)



import os.path
import tkinter as tk
from tkinter import ttk
import re
import tkutil
from exn.view.roll import Roll
from viewable import Viewable
from exn import constant


class Search(Viewable):
    def __init__(self, front):
        # target = "all" or "past"
        super().__init__()
        self._front = front
        self._results_frame = None
        self._results_front_frame = None
        self._search_strvar = tk.StringVar()
        self._found_status_strvar = tk.StringVar()
        self._list_status_strvar = tk.StringVar()
        self._deep_search_intvar = tk.IntVar()
        self._use_regex_intvar = tk.IntVar()
        self._marked_only_intvar = tk.IntVar()
        self._current_list_number = 0
        self._previously_mapped = False
        self._total_lists = 0
        self._roll = None
        self._search_entry = None
        self._matches = list()
        self._suggestion = None
        self._index_data = list()
        self._bookmarks = list()

    def _create_body(self, parent):
        return tk.Toplevel(parent, name="search")

    def _build(self):
        self.body.bind("<Escape>", lambda e: self._hide_window())
        self.body.title("")
        self.body.minsize(*constant.WINDOW_MINSIZE)
        tkutil.restore_size(self.body, name="exn_search",
                            default=constant.SEARCH_WINDOW_SIZE)
        # frames
        frame1 = tk.Frame(self.body)
        frame1.pack()
        frame2 = tk.Frame(self.body)
        frame2.pack(fill=tk.X, pady=(7,3))
        frame3 = tk.Frame(self.body)
        frame3.pack(pady=(7, 5))
        # SEARCH label
        label = tk.Label(frame1, name="title", text="Search")
        label.pack()
        # search entry
        self._search_entry = tk.Entry(frame2, name="search_field",
                                      textvariable=self._search_strvar)
        self._search_entry.pack(side=tk.LEFT, fill=tk.X, expand=True, padx=(3, 0))
        self._search_entry.bind("<Up>", lambda e: self._on_press_up())
        self._search_entry.bind("<Down>", lambda e: self._on_press_down())
        self._search_entry.focus_set()
        clear_entry = ttk.Button(frame2, text="x", style="clear2.TButton",
                                 command=self._on_clear_search)
        clear_entry.pack(side=tk.LEFT, padx=3)
        search_button = ttk.Button(frame2, name="search_button",
                                   style="blue.TButton",
                                   text="Search!", command=self._on_search)
        search_button.pack(side=tk.RIGHT, padx=(0, 3))
        # checkbuttons
        deep_search_checkbutton = ttk.Checkbutton(frame3, text="Deep Search",
                                                  style="toplevel.TCheckbutton",
                                                  variable=self._deep_search_intvar,
                                                  onvalue=1, offvalue=0,
                                                  command=self._on_toggle_deep_search)
        deep_search_checkbutton.pack(side=tk.LEFT, padx=2)
        regex_checkbutton = ttk.Checkbutton(frame3, text="Use RegEx",
                                            style="toplevel.TCheckbutton",
                                            variable=self._use_regex_intvar,
                                            onvalue=1, offvalue=0,
                                            command=self._on_toggle_use_regex)
        regex_checkbutton.pack(side=tk.LEFT, padx=2)
        marked_only_checkbutton = ttk.Checkbutton(frame3, text="Marked Only",
                                                  style="toplevel.TCheckbutton",
                                                  variable=self._marked_only_intvar,
                                                  onvalue=1, offvalue=0,
                                                  command=self._on_toggle_marked_only)
        marked_only_checkbutton.pack(side=tk.LEFT, padx=2)
        # center window
        editor = self._front.board.viewer.editor
        tkutil.align(self.body, parent=editor, side="n")
        tkutil.make_modal(self.body)

    def _on_map(self):
        self._index_data = list(self._front.board.viewer.pages.values())
        # suggestion
        dataset = list()
        for x in self._index_data:
            for tag in x.tags:
                tag = "#"+tag
                if tag in dataset:
                    continue
                dataset.append(tag)
        # TODO: uncomment the next two line, set the dropdown style, and fix Up/Down events conflicts
        #self._suggestion = Suggestion(self._search_entry, dataset=dataset)
        #self._suggestion.activate()
        self._search_entry.bind("<Return>", lambda e: self._on_search(), True)
        self._bookmarks = self._front.manager.dao.get_bookmarks()
        self._matches = self._index_data
        self._populate_roll(self._matches)

    def _on_destroy(self):
        tkutil.save_size(self.body, name="exn_search")

    def _install_header(self):
        header = tk.Frame(self.body)
        header.pack(fill=tk.X)
        # top side
        top_frame = tk.Frame(header)
        top_frame.pack(fill=tk.X)
        title_label = tk.Label(top_frame, text="Search")
        title_label.pack(side=tk.LEFT)
        # bottom side
        bottom_frame = tk.Frame(header)
        bottom_frame.pack()
        search_entry = tk.Entry(bottom_frame, textvariable=self._search_strvar)
        search_entry.pack(side=tk.LEFT)
        #search_entry.bind("<KeyRelease>", self._on_press_key, True)
        search_entry.bind("<Return>", self._on_click_search, True)
        #lite_search_checkbutton = tk.Checkbutton(frame, text="Lite Search")
        #lite_search_checkbutton.pack(side=tk.RIGHT)

    def _on_click_search(self, event):
        self._on_search()

    def _on_clear_search(self):
        self._search_strvar.set("")
        self._on_search()

    def _on_search(self):
        search_str = self._search_strvar.get()
        use_regex = True if self._use_regex_intvar.get() == 1 else False
        if self._deep_search_intvar.get() == 0:
            tags = [tag.strip("#") for tag in search_str.split()]
            results = self._lite_search(tags, use_regex)
        else:
            results = self._deep_search(search_str, use_regex)
        if self._marked_only_intvar.get() == 1:
            results = self._filter_marked_only(results)
        self._matches = results
        self._populate_roll(self._matches)

    def _on_toggle_deep_search(self):
        # TODO: reintegrate the commented lines
        self._on_search()
        if self._deep_search_intvar.get() == 1:
            #self._suggestion.deactivate()
            pass
        else:
            #self._suggestion.activate()
            pass

    def _on_toggle_use_regex(self):
        self._on_search()

    def _on_toggle_marked_only(self):
        if self._marked_only_intvar.get() == 1:
            self._matches = self._filter_marked_only(self._matches)
            self._populate_roll(self._matches)
        else:
            self._on_search()

    def _lite_search(self, tags, use_regex):
        results = list()
        for item in self._index_data:
            if not item.tags:
                continue
            matched = True
            if use_regex:  # if RegEx
                for tag in tags:
                    sub_matched = False
                    for x in item.tags:
                        if re.fullmatch(tag, x):
                            sub_matched = True
                            break
                    if not sub_matched:
                        matched = False
                        break
            else:  # No regex
                for tag in tags:
                    if tag not in item.tags:
                        matched = False
                        break
            if matched:
                results.append(item)
        return results

    def _deep_search(self, search_str, use_regex):
        results = list()
        dossier = self._front.manager.dossier
        for item in self._index_data:
            filename = item.filename
            path = os.path.join(dossier, *filename.split("/"))
            if not os.path.isfile(path):
                continue
            try:
                with open(path, "r") as file:
                    data = file.read()
                    if use_regex:
                        found = False
                        for _ in re.finditer(search_str, data):
                            found = True
                        if found:
                            results.append(item)
                    else:
                        if search_str in data:
                            results.append(item)
            except Exception as e:
                pass
        return results

    def _filter_marked_only(self, items):
        results = list()
        for item in items:
            if item.filename in self._bookmarks:
                results.append(item)
        return results

    def _on_press_up(self):
        self._roll.select(-1)

    def _on_press_down(self):
        self._roll.select(0)

    def _populate_roll(self, data):
        # Roll
        if self._roll and self._roll.body:
            self._roll.body.destroy()
        if not data:
            return
        self._roll = Roll(self._front, data, on_open=self._open_note,
                          on_close=self._hide_window,
                          on_leave_focus=self._on_leave_roll_focus)
        self._roll.build_pack(self.body, fill=tk.BOTH, expand=True)
        self._search_entry.focus_set()

    def _on_leave_roll_focus(self):
        self._search_entry.focus_set()
        return True

    def _open_note(self, filename):
        self.body.withdraw()
        self._front.open(filename)
        self._hide_window()

    def _hide_window(self):
        self.body.grab_release()
        self.body.withdraw()



import tkinter as tk
from tkinter import ttk
import tkutil
from viewable import Viewable
from exn.view import util


class Info(Viewable):
    def __init__(self, front):
        super().__init__()
        self._front = front
        self._title_strvar = tk.StringVar()
        self._page_strvar = tk.StringVar()
        self._filename_strvar = tk.StringVar()
        self._tags_strvar = tk.StringVar()
        self._mark_button = None

    def _create_body(self, parent):
        return tk.Toplevel(parent, name="info")

    def _build(self):
        self.body.bind("<Escape>", lambda e: self._on_close())
        self.body.title("")
        #self.body.minsize(*constant.WINDOW_MINSIZE)
        #tkutil.restore_size(self.body, name="exn_info",
        #                    default=constant.INFO_WINDOW_SIZE)
        # labels
        frame1 = tk.Frame(self.body)
        frame1.pack(pady=5)
        frame2 = tk.Frame(self.body, name="info_card")
        frame2.pack(fill=tk.BOTH, expand=True, padx=10, pady=10)
        frame3 = tk.Frame(self.body)
        frame3.pack(fill=tk.X, padx=3, pady=(15, 3))
        # Info
        info_label = tk.Label(frame1, name="title", text="Info")
        info_label.pack()
        # note title
        title_field = tk.Entry(frame2, name="title_field",
                               width=30,
                               textvariable=self._title_strvar,
                               state="disabled", takefocus=False,
                               cursor="arrow")
        title_field.pack(fill=tk.X, padx=5, pady=3)
        # page and filename frame
        frame = tk.Frame(frame2)
        frame.pack(fill=tk.X, padx=5, pady=(0, 3))
        # page number
        page_label = tk.Label(frame, name="page_label",
                              textvariable=self._page_strvar)
        page_label.pack(side=tk.LEFT)
        # note filename
        filename_field = tk.Entry(frame, name="filename_field",
                                  textvariable=self._filename_strvar,
                                  state="disabled", takefocus=False,
                                  cursor="arrow")
        filename_field.pack(side=tk.LEFT, fill=tk.X, expand=True, padx=(2, 0))
        # note tags
        tags_field = tk.Entry(frame2, name="tags_field",
                              textvariable=self._tags_strvar,
                              state="disabled", takefocus=False,
                              cursor="arrow")
        tags_field.pack(fill=tk.X, padx=5, pady=(0, 3))
        # footer
        close_button = ttk.Button(frame3, name="close", style="red.TButton",
                                  text="Close", command=self._on_close)
        close_button.pack(side=tk.LEFT)
        self._mark_button = ttk.Button(frame3, name="mark", text="Mark",
                                      command=self._on_click_mark)
        self._mark_button.pack(side=tk.RIGHT)
        copy_button = ttk.Button(frame3, name="copy", text="Copy",
                                command=self._on_copy)
        copy_button.pack(side=tk.RIGHT, padx=3)
        # center the window
        editor = self._front.board.viewer.editor
        tkutil.align(self.body, parent=editor, side="n")
        tkutil.make_modal(self.body)

    def _on_map(self):
        self._populate()
        filename = self._front.filename
        if self._front.manager.dao.is_bookmarked(filename):
            util.set_marked_style(self._mark_button, small=False)

    def _on_destroy(self):
        return
        #tkutil.save_size(self.body, name="exn_past")

    def _populate(self):
        title, page, filename, tags = self._get_data()
        self._title_strvar.set(title)
        self._page_strvar.set(page + ".")
        self._filename_strvar.set(filename)
        self._tags_strvar.set(tags)

    def _get_data(self):
        index = self._front.manager.index
        title = page = filename = tags = ""
        for item in index:
            if item.filename == self._front.filename:
                title = item.title
                page = str(item.page)
                filename = item.filename
                tags = ["#{}".format(tag) for tag in item.tags]
                tags = " ".join(tags)
                break
        return title, page, filename, tags

    def _on_copy(self):
        title = self._title_strvar.get()
        page = self._page_strvar.get()
        filename = self._filename_strvar.get()
        tags = self._tags_strvar.get()
        text = "{}\n{} {}\n{}".format(title, page, filename, tags)
        util.update_clipboard(text, self.body)

    def _on_click_mark(self):
        filename = self._front.filename
        if not filename:
            return
        util.on_click_mark_button(self._mark_button, filename,
                                  self._front.manager.dao, small=False)

    def _on_close(self):
        self.body.withdraw()
        self.body.destroy()



from collections import namedtuple


IndexEntry = namedtuple("IndexEntry", ["filename", "title", "tags"])

ExtensionContext = namedtuple("ExtensionContext", ["dossier", "app", "manager"])

SidebarContext = namedtuple("SidebarContext", ["dossier", "app", "manager", "path"])



import tkinter as tk
from tkinter import ttk
from viewable import Viewable
from exn.view.search import Search
from exn.view.toc import Toc
from exn.view.past import Past
from exn.view.info import Info
from exn.view.about import About
from exn.view.goto import Goto


class Top(Viewable):
    def __init__(self, front):
        super().__init__()
        self._front = front
        self._manager = front.manager
        self._title_strvar = tk.StringVar()
        self._title = None
        self._search_view = None

    def click_home(self):
        self._on_click_home()

    def click_toc(self):
        self._on_click_toc()

    def click_info(self):
        self._on_click_info()

    def click_goto(self):
        self._on_click_goto()

    def click_past(self):
        self._on_click_past()

    def click_search(self):
        self._on_click_search()

    def click_about(self):
        self._on_click_about()

    def _create_body(self, parent):
        return ttk.Frame(parent, name="top")

    def _build(self):
        self._install_menu_bar()
        #self._install_right_frame()

    def _install_left_frame(self):
        # left frame
        left_frame = ttk.Frame(self.body)
        #left_frame.grid(row=0, column=0, sticky="we")
        #left_frame.rowconfigure(0, weight=1)
        #left_frame.columnconfigure(1, weight=1)
        left_frame.pack(side=tk.LEFT, expand=True, fill=tk.BOTH,
                        pady=5)
        """
        # title label
        title_label = tk.Label(left_frame,
                               textvariable=self._title_label_strvar,
                               cursor="hand1")
        title_label.grid(row=0, column=0)
        title_label.bind("<Button-1>", self._on_click_title)
        title_label.bind("<Button-3>", self._on_right_click_title)
        """
        # title entry
        title_entry = tk.Entry(left_frame, textvariable=self._title_strvar,
                               takefocus=False, state="readonly",
                               cursor="hand1")
        #title_entry.grid(row=0, column=1, sticky="nswe")
        title_entry.pack(side=tk.LEFT, expand=True, fill=tk.BOTH, padx=1)
        title_entry.bind("<Button-1>", self._on_click_title)
        title_entry.bind("<Button-3>", self._on_right_click_title)
        title_entry.bind("<FocusIn>", lambda e: self._on_title_focus())

    def _install_menu_bar(self):
        # right frame
        frame = ttk.Frame(self.body)
        #right_frame.grid(row=0, column=1, sticky="e")
        frame.pack(pady=3)
        # about button
        about_button = ttk.Button(frame, text="About",
                                 takefocus=False,
                                 command=self._on_click_about)
        about_button.pack(side=tk.RIGHT, padx=(0, 3))
        # search button
        search_button = ttk.Button(frame, text="Search",
                                  takefocus=False,
                                  command=self._on_click_search)
        search_button.pack(side=tk.RIGHT, padx=(0, 3))
        # past history button
        past_button = ttk.Button(frame, text="Past",
                                takefocus=False,
                                command=self._on_click_past)
        past_button.pack(side=tk.RIGHT, padx=(0, 3))
        # goto button
        goto_button = ttk.Button(frame, text="Goto",
                                takefocus=False,
                                command=self._on_click_goto)
        goto_button.pack(side=tk.RIGHT, padx=(0, 3))
        # info button
        info_button = ttk.Button(frame, text="Info",
                                takefocus=False,
                                command=self._on_click_info)
        info_button.pack(side=tk.RIGHT, padx=(0, 3))
        # Table of Contents button
        toc_button = ttk.Button(frame, text="Toc",
                               takefocus=False,
                               command=self._on_click_toc)
        toc_button.pack(side=tk.RIGHT, padx=(0, 3))
        # home button
        home_button = ttk.Button(frame, text="Home",
                                takefocus=False,
                                command=self._on_click_home)
        home_button.pack(side=tk.RIGHT, padx=(0, 3))

    def _on_right_click_title(self, event):
        self._update_clipboard(self._title_strvar.get())
        # TODO
        #Toast(self.body, message="Copied !")

    def _on_click_home(self):
        pages = self._front.board.viewer.pages
        if not pages:
            return
        home = list(pages.values())[0]
        self._front.open(home.filename)

    def _on_click_past(self):
        past = Past(self._front)
        past.build(self.body)

    def _on_click_search(self):
        if self._search_view and self._search_view.body.winfo_exists():
            self._search_view.body.deiconify()
            self._search_view.body.grab_set()
            return
        self._search_view = Search(self._front)
        self._search_view.build(self.body)

    def _on_click_info(self):
        info = Info(self._front)
        info.build(self.body)

    def _on_click_goto(self):
        goto = Goto(self._front)
        goto.build(self.body)

    def _on_click_toc(self):
        viewer = self._front.board.viewer
        if not viewer:
            return
        if not viewer.sids:
            statusbar = viewer.statusbar
            statusbar.show("No sections !", duration=1500, delay=0)
            return
        toc = Toc(self._front)
        toc.build(self.body)

    def _on_click_about(self):
        goto = About(self._front)
        goto.build(self.body)

    def _on_title_focus(self):
        viewer = self._front.board.viewer
        if not viewer:
            return
        viewer.editor.focus_set()

    def _update_clipboard(self, text):
        self.body.clipboard_clear()
        self.body.clipboard_append(text)
